# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from typing import TYPE_CHECKING

import ray
from loguru import logger

from nemo_curator.backends.experimental.utils import get_available_cpu_gpu_resources

if TYPE_CHECKING:
    from ray.actor import ActorClass

    from nemo_curator.stages.base import ProcessingStage

    from .adapter import RayActorPoolStageAdapter
    from .raft_adapter import RayActorPoolRAFTAdapter

_LARGE_INT = 2**31 - 1


def calculate_optimal_actors_for_stage(
    stage: "ProcessingStage",
    num_tasks: int,
    reserved_cpus: float = 0.0,
    reserved_gpus: float = 0.0,
) -> int:
    """Calculate optimal number of actors for a stage."""
    # Get available resources (not total cluster resources)
    available_cpus, available_gpus = get_available_cpu_gpu_resources()
    # Reserve resources for system overhead
    available_cpus = max(0, available_cpus - reserved_cpus)
    available_gpus = max(0, available_gpus - reserved_gpus)

    # Calculate max actors based on CPU constraints
    max_actors_cpu = int(available_cpus // stage.resources.cpus) if stage.resources.cpus > 0 else _LARGE_INT

    # Calculate max actors based on GPU constraints
    max_actors_gpu = int(available_gpus // stage.resources.gpus) if stage.resources.gpus > 0 else _LARGE_INT

    # Take the minimum constraint
    max_actors_resources = min(max_actors_cpu, max_actors_gpu)

    # Ensure we don't create more actors than configured maximum
    max_actors_resources = min(max_actors_resources, stage.num_workers() or _LARGE_INT)

    if max_actors_resources == 0:
        msg = f"No resources available for stage {stage.name}."
        raise ValueError(msg)

    number_of_batches = (
        math.ceil(num_tasks / stage.batch_size) if stage.batch_size is not None and stage.batch_size > 0 else num_tasks
    )
    # Don't create more actors than batches of work
    optimal_actors = min(number_of_batches, max_actors_resources)

    # Ensure at least 1 actor if we have tasks
    optimal_actors = max(1, optimal_actors) if num_tasks > 0 else 0

    logger.info(f"    Resource calculation: CPU limit={max_actors_cpu}, GPU limit={max_actors_gpu}")
    logger.info(f"    Available: {available_cpus} CPUs, {available_gpus} GPUs")
    logger.info(f"    Stage requirements: {stage.resources.cpus} CPUs, {stage.resources.gpus} GPUs")

    return optimal_actors


def create_named_ray_actor_pool_stage_adapter(
    stage: "ProcessingStage",
    cls: type["RayActorPoolStageAdapter"] | type["RayActorPoolRAFTAdapter"],
) -> "ActorClass[RayActorPoolStageAdapter | RayActorPoolRAFTAdapter]":
    """Create a named RayActorPoolStageAdapter or RayActorPoolRAFTAdapter.

    This function creates a dynamic subclass of the given adapter class,
    named after the stage's class name. This ensures that when Ray calls
    type(adapter).__name__, it returns the original stage's class name rather
    than 'RayActorPoolStageAdapter' or 'RayActorPoolRAFTAdapter'.

    Args:
        stage (ProcessingStage): ProcessingStage to adapt
        cls (type): The adapter class to inherit from

    Returns:
        ActorClass: A ray.remote decorated class that can be used to create actors
    """
    # Get the original stage's class name
    original_class_name = type(stage).__name__

    # Create a dynamic subclass with the original name
    DynamicAdapter = type(  # noqa: N806
        original_class_name,  # Use the original stage's name
        (cls,),  # Inherit from the adapter class
        {
            "__module__": cls.__module__,  # Keep the same module
        },
    )

    # Return the ray.remote decorated class
    return ray.remote(DynamicAdapter)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

import ray
from loguru import logger

from nemo_curator.backends.base import BaseStageAdapter
from nemo_curator.backends.experimental.utils import get_worker_metadata_and_node_id
from nemo_curator.stages.base import ProcessingStage

if TYPE_CHECKING:
    from nemo_curator.backends.base import WorkerMetadata


class RayActorPoolRAFTAdapter(BaseStageAdapter):
    """RAFT Actor adapter for Ray Actor Pool backend.

    This adapter extends RayActorPoolStageAdapter and adds RAFT capabilities
    to enable distributed processing with RAFT communication.
    """

    def __init__(
        self, stage: ProcessingStage, index: int, pool_size: int, session_id: bytes, actor_name_prefix: str = "RAFT"
    ):
        """Initialize the RAFT adapter.

        Args:
            stage: The processing stage to wrap
            index: The index of this actor in the pool
            pool_size: Total number of actors in the pool
            session_id: Unique session identifier
            actor_name_prefix: Prefix for actor names
        """
        # Initialize the base stage adapter first
        super().__init__(stage)

        # Get runtime context for worker metadata (copied from RayActorPoolStageAdapter)
        node_info, worker_metadata = get_worker_metadata_and_node_id()

        # Create WorkerMetadata with actor information
        self.worker_metadata = worker_metadata
        self.node_info = node_info

        self._batch_size = self.stage.batch_size
        if self._batch_size is None:
            logger.warning(f"batch size not set for stage {self.stage}. Setting it to 1.")
            self._batch_size = 1

        # Initialize RAFT-specific attributes
        self._index = index
        self._actor_name_prefix = actor_name_prefix
        self._name = f"{self._actor_name_prefix}Actor-{self._index}"
        self._pool_size = pool_size
        self._is_root = not index
        self.session_id = session_id

        # Initialize RAFT communication
        from nemo_curator.backends.internal.raft.ray_comms import Comms

        self.cb = Comms(verbose=True, nccl_root_location="ray-actor")
        self.cb.init()
        self.unique_id = self.cb.uniqueId
        self.root_unique_id = self.unique_id if self._is_root else None

        logger.debug(f"Initialized RAFT adapter {self._name} (index={index}, is_root={self._is_root})")

    def get_batch_size(self) -> int:
        """Get the batch size for this stage."""
        return self._batch_size

    def setup_on_node(self) -> None:
        """Setup method for Ray actors.

        Note: This method is not used in the current implementation since we use
        the Ray Data pattern of calling setup_on_node before actor creation.
        """
        super().setup_on_node(self.node_info, self.worker_metadata)

    def broadcast_root_unique_id(self) -> None:
        """Broadcast the root unique ID to all actors.

        This method should only be called by the root actor.
        """
        if self._is_root:
            actor_handles = [
                ray.get_actor(name=f"{self._actor_name_prefix}Actor-{i}", namespace=None)
                for i in range(1, self._pool_size)
            ]
            futures = [actor.set_root_unique_id.remote(self.root_unique_id) for actor in actor_handles]

            # Block until all futures complete
            ray.get(futures)
        else:
            msg = "This method should only be called by the root"
            raise RuntimeError(msg)

    def set_root_unique_id(self, root_unique_id: int) -> None:
        """Set the root unique ID.

        Parameters
        ----------
        root_unique_id : int
            The root unique ID.
        """
        logger.debug(f"{self._name}: set_root_unique_id")
        if self.root_unique_id is None:
            self.root_unique_id = root_unique_id

    def _setup_nccl(self) -> None:
        """Setup NCCL communicator."""
        from raft_dask.common.nccl import nccl

        self._nccl = nccl()
        self._nccl.init(self._pool_size, self.root_unique_id, self._index)

    def _setup_raft(self) -> None:
        """Setup RAFT."""
        from pylibraft.common.handle import Handle
        from raft_dask.common.comms_utils import inject_comms_on_handle_coll_only

        self._raft_handle = Handle(n_streams=0)
        inject_comms_on_handle_coll_only(self._raft_handle, self._nccl, self._pool_size, self._index, verbose=True)

    def setup(self, worker_metadata: "WorkerMetadata | None" = None) -> None:
        """Setup the RAFT actor.

        This method should be called after the root unique ID has been broadcast.
        """
        if self.root_unique_id is None:
            msg = "The unique ID of root is not set. Make sure `broadcast_root_unique_id` "
            "runs on the root before calling this method."
            raise RuntimeError(msg)

        try:
            self._setup_nccl()
            self._setup_raft()
            # Set the RAFT handle on the stage so it can access it
            self.stage._raft_handle = self._raft_handle
            self.stage._actor_pool_size = self._pool_size
            # This calls the stage's setup method
            super().setup(worker_metadata)
        except Exception as e:
            logger.error(f"An error occurred while setting up {self._name}: {e}.")
            raise

    def teardown(self) -> None:
        super().teardown()



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal

import cupy as cp
import numpy as np

from nemo_curator.backends.base import WorkerMetadata
from nemo_curator.stages.base import CompositeStage, ProcessingStage
from nemo_curator.stages.deduplication.io_utils import DeduplicationIO
from nemo_curator.stages.file_partitioning import FilePartitioningStage
from nemo_curator.stages.resources import Resources
from nemo_curator.stages.text.embedders.utils import create_list_series_from_1d_or_2d_ar
from nemo_curator.tasks import FileGroupTask, _EmptyTask
from nemo_curator.utils.file_utils import FILETYPE_TO_DEFAULT_EXTENSIONS, check_disallowed_kwargs

from .utils import break_parquet_partition_into_groups, get_array_from_df

if TYPE_CHECKING:
    import cudf

import time

import torch
from loguru import logger

# Column names
L2_DIST_TO_CENT_COL = "l2_dist_to_cent"
COSINE_DIST_TO_CENT_COL = "cosine_dist_to_cent"


class KMeansReadFitWriteStage(ProcessingStage[FileGroupTask, _EmptyTask], DeduplicationIO):
    """KMeans clustering stage that requires RAFT for distributed processing."""

    def __init__(  # noqa: PLR0913
        self,
        id_field: str,
        embedding_field: str,
        output_path: str,
        filetype: Literal["parquet", "jsonl"],
        # KMeans args
        n_clusters: int,
        metadata_fields: list[str] | None = None,
        embedding_dim: int | None = None,
        verbose: bool = False,
        max_iter: int = 300,
        tol: float = 1e-4,
        random_state: int = 42,
        init: Literal["k-means||", "random"] | np.ndarray = "k-means||",
        n_init: int | Literal["auto"] = 1,
        oversampling_factor: float = 2.0,
        max_samples_per_batch: int = 1 << 15,
        # I/O args
        read_kwargs: dict[dict] | None = None,
        write_kwargs: dict[dict] | None = None,
    ):
        """KMeans clustering stage that requires RAFT for distributed processing.

        Args:
            id_field (str): The column name of the id column.
            embedding_field (str): The column name of the embedding column.
            output_path (str): The path to the output directory.
            n_clusters (int): The number of clusters to create.
            metadata_fields (list[str] | None): The columns to keep in the output. These columns can be used later to prioritize deduplication.
            embedding_dim (int | None): The dimension of the embedding. This helps us read data into smaller chunks.
            verbose (bool): Whether to print verbose output.
            max_iter (int): The maximum number of iterations to run.
            tol (float): Tolerance for stopping criteria of the kmeans algorithm.
            random_state (int): Seed for the random number generator. Unseeded by default. Does not currently fully guarantee the exact same results.
            init (Literal["k-means||", "random"] | np.ndarray): 'scalable-k-means++' or 'k-means||': Uses fast and stable scalable kmeans++ initialization. 'random': Choose 'n_cluster' observations (rows) at random from data for the initial centroids. If an ndarray is passed, it should be of shape (n_clusters, n_features) and gives the initial centers.
            n_init (int | Literal["auto"]): Number of times the k-means algorithm will be run with different centroid seeds. The final results will be the best output of n_init consecutive runs in terms of inertia.
            oversampling_factor (float): The amount of points to sample in scalable k-means++ initialization for potential centroids. Increasing this value can lead to better initial centroids at the cost of memory. The total number of centroids sampled in scalable k-means++ is oversampling_factor * n_clusters * 8.
            max_samples_per_batch (int): The number of data samples to use for batches of the pairwise distance computation. This computation is done throughout both fit predict. The default should suit most cases. The total number of elements in the batched pairwise distance computation is max_samples_per_batch * n_clusters. It might become necessary to lower this number when n_clusters becomes prohibitively large.
            read_kwargs (dict[dict]): Keyword arguments for the read stage.
            write_kwargs (dict[dict]): Keyword arguments for the write stage.
        """
        self.id_field = id_field
        self.embedding_field = embedding_field
        self.output_path = output_path
        self.filetype = filetype
        self.n_clusters = n_clusters
        self.metadata_fields = metadata_fields if metadata_fields is not None else []
        self.embedding_dim = embedding_dim
        self.verbose = verbose
        self.max_iter = max_iter
        self.tol = tol
        self.random_state = random_state
        self.init = init
        self.n_init = n_init
        self.oversampling_factor = oversampling_factor
        self.max_samples_per_batch = max_samples_per_batch

        self.read_kwargs = read_kwargs.copy() if read_kwargs is not None else {}
        self.write_kwargs = write_kwargs.copy() if write_kwargs is not None else {}

        check_disallowed_kwargs(self.read_kwargs, ["columns", "assign_id"])
        check_disallowed_kwargs(self.write_kwargs, ["partition_file_name", "partition_cols", "index"])

        self.input_storage_options = self.read_kwargs.pop("storage_options", None)
        self.output_storage_options = self.write_kwargs.pop("storage_options", None)

        self._name = "KMeansStage"
        self._resources = Resources(cpus=1.0, gpus=1.0)

    def process(self, task: FileGroupTask) -> _EmptyTask:
        msg = "KMeansReadFitWriteStage does not support single-task processing"
        raise NotImplementedError(msg)

    def process_batch(self, tasks: list[FileGroupTask]) -> list[_EmptyTask]:
        """Process a batch of FileGroupTasks using distributed RAFT KMeans.

        In RAFT mode, each actor processes its assigned tasks, but the KMeans model
        is trained cooperatively across all actors using RAFT communication.

        This method:
        1. Reads data from this actor's assigned tasks
        2. Breaks data into subgroups to avoid cudf row limits
        3. Fits distributed KMeans model (coordinates with other actors via RAFT)
        4. Assigns cluster centroids back to each subgroup
        5. Writes the results for each subgroup
        """

        if not tasks:
            return []

        # Collect all files from all tasks
        all_files = [file for task in tasks for file in task.data]

        # Break files into subgroups to avoid cudf row limits
        if self.filetype == "parquet":
            groups = break_parquet_partition_into_groups(all_files, embedding_dim=self.embedding_dim)
        elif self.filetype == "jsonl":
            # For JSONL files, just group all files together since we can't easily estimate size
            groups = [all_files]
        else:
            msg = f"Unsupported filetype: {self.filetype}. Only jsonl and parquet are supported."
            raise ValueError(msg)

        # Read each subgroup independently
        t0 = time.perf_counter()
        all_dfs, embeddings_arrays = [], []

        for group in groups:
            # Read all files in this group
            if self.filetype == "parquet":
                df = self.read_parquet(
                    group,
                    columns=[self.id_field, self.embedding_field, *self.metadata_fields],
                    storage_options=self.input_storage_options,
                    assign_id=False,
                    **self.read_kwargs,
                )
            elif self.filetype == "jsonl":
                df = self.read_jsonl(
                    group,
                    columns=[self.id_field, self.embedding_field, *self.metadata_fields],
                    storage_options=self.input_storage_options,
                    assign_id=False,
                    **self.read_kwargs,
                )
            else:
                msg = f"Unsupported data type: {self.filetype}"
                raise ValueError(msg)

            # Normalize the embeddings
            df = self.normalize_embeddings_col_in_df(df, self.embedding_field)

            # Convert embeddings to cupy array to avoid cudf row limits
            embeddings_array = get_array_from_df(df, self.embedding_field)

            # Maintain a list of DataFrames and embeddings arrays for later use
            all_dfs.append(df)
            embeddings_arrays.append(embeddings_array)

        t1 = time.perf_counter()
        logger.debug(f"Read time: {(t1 - t0):.2f} seconds")
        # Fit the model cooperatively across actors, then predict on local data
        concatenated_embeddings = cp.concatenate(embeddings_arrays, axis=0)
        self.kmeans.fit(concatenated_embeddings, sample_weight=None)
        labels = self.kmeans.predict(concatenated_embeddings).astype(cp.int32)

        t2 = time.perf_counter()
        logger.info(f"KMeans fit+predict time: {(t2 - t1):.2f} seconds")

        results = []
        num_rows_seen = 0
        # Assign labels back to DataFrame and write results
        for i, df in enumerate(all_dfs):
            end_idx = num_rows_seen + len(df)
            df["centroid"] = labels[num_rows_seen:end_idx]
            num_rows_seen = end_idx
            # Assign distances using the fitted cluster centers
            df = self._assign_distances(df, self.embedding_field, self.kmeans.cluster_centers_)  # noqa: PLW2901

            output_filename = f"{tasks[0]._uuid}_{i}"
            # Write results for this subgroup
            self.write_parquet(
                df,
                self.output_path,
                partition_file_name=f"{output_filename}.parquet",
                partition_cols=["centroid"],
                index=False,
                storage_options=self.output_storage_options,
                **self.write_kwargs,
            )

            # Create result task for this subgroup
            results.append(
                _EmptyTask(
                    task_id=output_filename,
                    dataset_name=f"kmeans_group_{i}",
                    _metadata=None,
                    _stage_perf=[],
                    data=None,
                )
            )
        t3 = time.perf_counter()
        logger.info(f"Write time: {(t3 - t2):.2f} seconds")

        return results

    def setup(self, _: WorkerMetadata | None = None) -> None:
        from cuml.cluster.kmeans_mg import KMeansMG as cumlKMeans

        if not hasattr(self, "_raft_handle"):
            msg = "RAFT handle not found. Make sure the stage is initialized with RAFT"
            raise ValueError(msg)

        self.kmeans = cumlKMeans(
            handle=self._raft_handle,
            output_type="cupy",
            init=self.init,
            n_clusters=self.n_clusters,
            max_iter=self.max_iter,
            tol=self.tol,
            verbose=self.verbose,
            n_init=self.n_init,
            oversampling_factor=self.oversampling_factor,
            max_samples_per_batch=self.max_samples_per_batch,
            convert_dtype=False,
        )

    @staticmethod
    def normalize_embeddings_col_in_df(df: "cudf.DataFrame", embedding_col: str) -> "cudf.DataFrame":
        tensor = torch.Tensor(get_array_from_df(df, embedding_col))
        normalized_tensor = tensor / torch.norm(tensor, dim=1, keepdim=True)
        df[embedding_col] = create_list_series_from_1d_or_2d_ar(cp.asarray(normalized_tensor), index=df.index)
        return df

    @staticmethod
    def _assign_distances(df: "cudf.DataFrame", embedding_col: str, centroids: "cp.ndarray") -> "cudf.DataFrame":
        """
        Computes the L2 distance to nearest centroid to each embedding in the DataFrame.
        Embeddings are normalized. For cosine we'll need to normalize the centroids as well.
        """
        normalized_embeddings = get_array_from_df(df, embedding_col)
        # We normalize the centroids as well for cosine distance
        normalized_centroids = centroids / cp.linalg.norm(centroids, axis=1, keepdims=True)

        df[L2_DIST_TO_CENT_COL] = cp.sqrt(
            cp.sum((normalized_embeddings - centroids[df["centroid"].values]) ** 2, axis=1)
        )
        df[COSINE_DIST_TO_CENT_COL] = 1 - (
            cp.sum(
                normalized_embeddings * normalized_centroids[df["centroid"].values],
                axis=1,
            )
        )
        return df

    def ray_stage_spec(self) -> dict[str, Any]:
        return {
            "is_raft_actor": True,
        }


@dataclass
class KMeansStage(CompositeStage[_EmptyTask, _EmptyTask]):
    """KMeans clustering stage that requires RAFT for distributed processing."""

    n_clusters: int
    id_field: str
    embedding_field: str
    input_path: str | list[str]
    output_path: str
    metadata_fields: list[str] | None = None
    verbose: bool = False
    embedding_dim: int | None = None
    # I/O args
    input_filetype: Literal["jsonl", "parquet"] = "parquet"
    input_file_extensions: list[str] | None = None
    read_kwargs: dict[dict] | None = None
    write_kwargs: dict[dict] | None = None
    # KMeans args
    max_iter: int = 300
    tol: float = 1e-4
    random_state: int = 42
    init: Literal["k-means||", "random"] | np.ndarray = "k-means||"
    n_init: int | Literal["auto"] = 1
    oversampling_factor: float = 2.0
    max_samples_per_batch: int = 1 << 15
    """KMeans clustering stage that requires RAFT for distributed processing.

    Args:
        n_clusters (int): The number of clusters to create.
        id_field (str): The column name of the id column.
        embedding_field (str): The column name of the embedding column.
        input_path (str | list[str]): The path to the input directory.
        output_path (str): The path to the output directory.
        metadata_fields (list[str] | None): The columns to keep in the output. These columns can be used later to prioritize deduplication.
        verbose (bool): Whether to print verbose output.
        embedding_dim (int | None): The dimension of the embedding. This helps us read data into smaller chunks.
        input_filetype (Literal["jsonl", "parquet"]): The type of the input file
        read_kwargs (dict[dict]): Keyword arguments for the read stage.
        write_kwargs (dict[dict]): Keyword arguments for the write stage.
        max_iter (int): The maximum number of iterations to run.
        tol (float): Tolerance for stopping criteria of the kmeans algorithm.
        random_state (int): Seed for the random number generator. Unseeded by default. Does not currently fully guarantee the exact same results.
        init (Literal["k-means||", "random"] | np.ndarray): 'scalable-k-means++' or 'k-means||': Uses fast and stable scalable kmeans++ initialization. 'random': Choose 'n_cluster' observations (rows) at random from data for the initial centroids. If an ndarray is passed, it should be of shape (n_clusters, n_features) and gives the initial centers.
        n_init (int | Literal["auto"]): Number of times the k-means algorithm will be run with different centroid seeds. The final results will be the best output of n_init consecutive runs in terms of inertia.
        oversampling_factor (float): The amount of points to sample in scalable k-means++ initialization for potential centroids. Increasing this value can lead to better initial centroids at the cost of memory. The total number of centroids sampled in scalable k-means++ is oversampling_factor * n_clusters * 8.
        max_samples_per_batch (int): The number of data samples to use for batches of the pairwise distance computation. This computation is done throughout both fit predict. The default should suit most cases. The total number of elements in the batched pairwise distance computation is max_samples_per_batch * n_clusters. It might become necessary to lower this number when n_clusters becomes prohibitively large.
    """

    def __post_init__(self):
        """Initialize parent class after dataclass initialization."""
        super().__init__()

    def decompose(self) -> list[ProcessingStage]:
        # Set default file extensions based on input_filetype if not provided
        file_extensions = self.input_file_extensions or FILETYPE_TO_DEFAULT_EXTENSIONS.get(self.input_filetype, [])
        if not file_extensions:
            msg = f"Unsupported filetype: {self.input_filetype}"
            raise ValueError(msg)

        return [
            FilePartitioningStage(
                file_paths=self.input_path,
                file_extensions=file_extensions,
                files_per_partition=1,  # We set this to one, and then the RaftActor will break it up into smaller groups
                storage_options=self.read_kwargs.get("storage_options") if self.read_kwargs is not None else None,
            ),
            KMeansReadFitWriteStage(
                id_field=self.id_field,
                embedding_field=self.embedding_field,
                output_path=self.output_path,
                filetype=self.input_filetype,
                n_clusters=self.n_clusters,
                metadata_fields=self.metadata_fields,
                verbose=self.verbose,
                embedding_dim=self.embedding_dim,
                max_iter=self.max_iter,
                tol=self.tol,
                random_state=self.random_state,
                init=self.init,
                n_init=self.n_init,
                oversampling_factor=self.oversampling_factor,
                max_samples_per_batch=self.max_samples_per_batch,
                read_kwargs=self.read_kwargs,
                write_kwargs=self.write_kwargs,
            ),
        ]



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import subprocess
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

from loguru import logger

from nemo_curator.stages.base import ProcessingStage
from nemo_curator.stages.resources import Resources
from nemo_curator.tasks import FileGroupTask


class DocumentDownloader(ABC):
    """Abstract base class for document downloaders."""

    def __init__(self, download_dir: str, verbose: bool = False):
        """Initialize the downloader.

        Args:
            download_dir: Directory to store downloaded files
            verbose: If True, logs detailed download information
        """
        self._download_dir = download_dir
        self._verbose = verbose
        os.makedirs(download_dir, exist_ok=True)

    def _check_s5cmd_installed(self) -> bool:
        """Check if s5cmd is installed."""
        try:
            subprocess.run(["s5cmd", "version"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, check=False)  # noqa: S603, S607
        except FileNotFoundError:
            return False
        else:
            return True

    @abstractmethod
    def _get_output_filename(self, url: str) -> str:
        """Generate output filename from URL.

        Args:
            url: URL to download

        Returns:
            Output filename (without directory path)
        """
        ...

    @abstractmethod
    def _download_to_path(self, url: str, path: str) -> tuple[bool, str | None]:
        """Download URL to specified path.

        Args:
            url: URL to download
            path: Local path to save file

        Returns:
            Tuple of (success, error_message). If success is True, error_message should be None.
            If success is False, error_message should contain the error details.
        """
        ...

    def download(self, url: str) -> str | None:
        """Download a document from URL with temporary file handling.

        Downloads file to temporary location then atomically moves to final path.
        Checks for existing file to avoid re-downloading. Supports resumable downloads.
        Args:
            url: URL to download

        Returns:
            Path to downloaded file, or None if download failed
        """
        # Generate output filename
        output_name = self._get_output_filename(url)
        output_file = os.path.join(self._download_dir, output_name)
        temp_file = output_file + ".tmp"

        # If final file exists and is non-empty, assume it's complete
        if os.path.exists(output_file) and os.path.getsize(output_file) > 0:
            if self._verbose:
                logger.info(f"File: {output_file} exists. Not downloading")
            return output_file

        # Download to temporary file
        success, error_message = self._download_to_path(url, temp_file)

        if success:
            # Download successful, atomically move temp file to final location
            os.rename(temp_file, output_file)
            if self._verbose:
                file_size = os.path.getsize(output_file)
                logger.info(f"Successfully downloaded to {output_file} ({file_size} bytes)")
            return output_file
        else:
            # Download failed
            logger.error(f"Failed to download to {output_file}: {error_message}")
            return None

    def num_workers_per_node(self) -> int | None:
        """Number of workers per node for Downloading. This is sometimes needed to ensure we are not overloading the network.

        Returns:
            Number of workers per node, or None if there is no limit and we can download as fast as possible
        """
        return None


@dataclass
class DocumentDownloadStage(ProcessingStage[FileGroupTask, FileGroupTask]):
    """Stage that downloads files from URLs to local storage.

    Takes a FileGroupTask with URLs and returns a FileGroupTask with local file paths.
    This allows the download step to scale independently from iteration/extraction.
    """

    _resources = Resources(cpus=0.5)
    downloader: DocumentDownloader
    _batch_size = None

    def __post_init__(self):
        self._name = f"download_{self.downloader.__class__.__name__.lower()}"

    def inputs(self) -> tuple[list[str], list[str]]:
        """Define input requirements - expects FileGroupTask with URLs."""
        return (["data"], [])

    def outputs(self) -> tuple[list[str], list[str]]:
        """Define output - produces FileGroupTask with local paths."""
        return (["data"], [])

    def process(self, task: FileGroupTask) -> FileGroupTask:
        """Download URLs to local files.

        Args:
            task (FileGroupTask): Task containing URLs to download

        Returns:
            FileGroupTask: Task containing local file paths
        """
        local_files = []

        for url in task.data:
            downloaded_file = self.downloader.download(url)
            if downloaded_file:
                local_files.append(downloaded_file)

        return FileGroupTask(
            task_id=task.task_id,
            dataset_name=task.dataset_name,
            data=local_files,
            _metadata={
                **task._metadata,
                "source_files": local_files,  # Add downloaded files for deterministic naming during write stage
            },
            _stage_perf=task._stage_perf,
        )

    def xenna_stage_spec(self) -> dict[str, Any]:
        return {
            "num_workers_per_node": self.downloader.num_workers_per_node(),
        }



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

import pandas as pd
from loguru import logger

from nemo_curator.stages.base import ProcessingStage
from nemo_curator.tasks import DocumentBatch
from nemo_curator.utils.column_utils import resolve_filename_column


class DocumentExtractor(ABC):
    """Abstract base class for document extractors.

    Takes a record dict and returns processed record dict or None to skip.
    Can transform any fields in the input dict.
    """

    @abstractmethod
    def extract(self, record: dict[str, str]) -> dict[str, Any] | None:
        """Extract/transform a record dict into final record dict."""
        ...

    @abstractmethod
    def input_columns(self) -> list[str]:
        """Define input columns - produces DocumentBatch with records."""
        ...

    @abstractmethod
    def output_columns(self) -> list[str]:
        """Define output columns - produces DocumentBatch with records."""
        ...


@dataclass
class DocumentExtractStage(ProcessingStage[DocumentBatch, DocumentBatch]):
    """Stage that extracts structured content from raw records.

    Takes DocumentBatch with raw content and produces DocumentBatch with extracted content.
    This is for cases where iteration and extraction are separate steps.
    """

    extractor: DocumentExtractor
    add_filename_column: bool | str = True

    def __post_init__(self):
        """Initialize the stage."""
        self.filename_col = resolve_filename_column(self.add_filename_column)
        self._name = f"extract_{self.extractor.__class__.__name__.lower()}"

    def inputs(self) -> tuple[list[str], list[str]]:
        """Define input requirements - expects DocumentBatch with dict records."""
        return (["data"], self.extractor.input_columns() + ([self.filename_col] if self.add_filename_column else []))  # type: ignore[reportReturnType]

    def outputs(self) -> tuple[list[str], list[str]]:
        """Define output - produces DocumentBatch with processed records."""
        return (["data"], self.extractor.output_columns() + ([self.filename_col] if self.add_filename_column else []))  # type: ignore[reportReturnType]

    def process(self, task: DocumentBatch) -> DocumentBatch:
        """Extract structured content from raw records.

        Args:
            task (DocumentBatch): Batch containing records

        Returns:
            DocumentBatch: Batch containing extracted records
        """
        extracted_records = []

        for _, row in task.data.iterrows():
            # Convert pandas Series to dict
            record_dict = row.to_dict()

            # Extract structured content
            extracted = self.extractor.extract(record_dict)
            if extracted is not None:
                if self.add_filename_column:
                    if self.filename_col in extracted:
                        msg = f"Since add_filename_col is specified, we'll overwrite ({self.filename_col}) from the input data."
                        logger.warning(msg)

                    extracted[self.filename_col] = record_dict[self.filename_col]  # type: ignore[reportReturnType]
                extracted_records.append(extracted)

        # Convert to DataFrame
        df = pd.DataFrame(extracted_records)

        return DocumentBatch(
            task_id=task.task_id,
            dataset_name=task.dataset_name,
            data=df,
            _metadata={
                **task._metadata,
            },
            _stage_perf=task._stage_perf,
        )



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Literal

from nemo_curator.backends.experimental.utils import RayStageSpecKeys
from nemo_curator.stages.base import ProcessingStage
from nemo_curator.stages.deduplication.shuffle_utils.rapidsmpf_shuffler import BulkRapidsMPFShuffler
from nemo_curator.stages.resources import Resources
from nemo_curator.tasks import FileGroupTask
from nemo_curator.utils.file_utils import create_or_overwrite_dir


class ShuffleStage(ProcessingStage[FileGroupTask, FileGroupTask]):
    """
    Stage that performs generic shuffling on specified columns from a FileGroupTask.
    This stage uses the BulkRapidsMPFShuffler with cuDF I/O for efficient GPU-based shuffling.

    Parameters
    ----------
    shuffle_on
        List of column names to shuffle on.
    total_nparts
        Total number of output partitions. If None, will be set automatically by the executor.
    output_path
        Path to write output files.
    read_kwargs
        Keyword arguments for cudf.read_parquet method.
    write_kwargs
        Keyword arguments for cudf.to_parquet method.
    rmm_pool_size
        Size of the RMM GPU memory pool in bytes.
        If "auto", the memory pool is set to 90% of the free GPU memory.
        If None, the memory pool is set to 50% of the free GPU memory that can expand if needed.
    spill_memory_limit
        Device memory limit in bytes for spilling to host.
        If "auto", the limit is set to 80% of the RMM pool size.
        If None spilling is disabled.
    enable_statistics
        Whether the underlying rapidsmpf shuffler should collect shuffle statistics.
    """

    _name = "ShuffleStage"
    _resources = Resources(gpus=1.0)

    # Use BulkRapidsMPFShuffler directly
    actor_class = BulkRapidsMPFShuffler

    def __init__(  # noqa: PLR0913
        self,
        shuffle_on: list[str],
        total_nparts: int | None = None,
        output_path: str = "./",
        read_kwargs: dict[str, Any] | None = None,
        write_kwargs: dict[str, Any] | None = None,
        rmm_pool_size: int | Literal["auto"] | None = "auto",
        spill_memory_limit: int | Literal["auto"] | None = "auto",
        enable_statistics: bool = False,
    ):
        super().__init__()

        # Initialize instance variables
        self.shuffle_on = shuffle_on
        self.total_nparts = total_nparts
        self.output_path = output_path
        self.rmm_pool_size = rmm_pool_size
        self.spill_memory_limit = spill_memory_limit
        self.enable_statistics = enable_statistics

        self.read_kwargs = read_kwargs if read_kwargs is not None else {}
        self.write_kwargs = write_kwargs if write_kwargs is not None else {}

        self.actor_kwargs = {
            "shuffle_on": self.shuffle_on,
            "total_nparts": self.total_nparts,  # Can be None, executor will set it
            "output_path": self.output_path,
            "rmm_pool_size": self.rmm_pool_size,
            "spill_memory_limit": self.spill_memory_limit,
            "enable_statistics": self.enable_statistics,
            "read_kwargs": self.read_kwargs,
            "write_kwargs": self.write_kwargs,
        }
        # Handle output path
        create_or_overwrite_dir(self.output_path, storage_options=self.write_kwargs.get("storage_options", {}))

    def process(self, task: FileGroupTask) -> FileGroupTask:
        """Not implemented for actor-based stages."""
        msg = "ShufflerStage does not support the process method. Use with an actor-based executor."
        raise NotImplementedError(msg)

    def ray_stage_spec(self) -> dict[str, Any]:
        """Ray stage specification for this stage."""
        return {
            RayStageSpecKeys.IS_SHUFFLE_STAGE: True,
        }

    def _check_actor_obj(self) -> None:
        """Verify the actor object is properly initialized."""
        if not hasattr(self, "_actor_obj") or not isinstance(self._actor_obj, self.actor_class):
            msg = (
                "Actor object not initialized. This might be because an incorrect executor "
                "was used or it failed to setup the stage properly."
            )
            raise RuntimeError(msg)

    def read_and_insert(self, task: FileGroupTask) -> FileGroupTask:
        """Read files and insert into shuffler."""
        self._check_actor_obj()
        self.output_columns = self._actor_obj.read_and_insert(task.data)
        self.dataset_name = task.dataset_name
        return task

    def insert_finished(self) -> None:
        self._check_actor_obj()
        self._actor_obj.insert_finished()

    def extract_and_write(self) -> list[FileGroupTask]:
        self._check_actor_obj()
        partition_paths = self._actor_obj.extract_and_write(column_names=self.output_columns)
        return [
            FileGroupTask(
                task_id=partition_id,
                dataset_name=self.dataset_name + f"{self.name}",
                data=[path],
                _metadata={
                    "partition_index": partition_id,
                    "total_partitions": len(partition_paths),
                    "output_columns": self.output_columns,
                },
            )
            for partition_id, path in partition_paths
        ]

    def teardown(self) -> None:
        self._check_actor_obj()
        self._actor_obj.cleanup()



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Any

import ray
from loguru import logger
from ray.data import DataContext, Dataset

from nemo_curator.backends.base import BaseExecutor
from nemo_curator.backends.experimental.utils import execute_setup_on_node
from nemo_curator.backends.utils import register_loguru_serializer
from nemo_curator.tasks import EmptyTask, Task

from .adapter import RayDataStageAdapter

if TYPE_CHECKING:
    from nemo_curator.stages.base import ProcessingStage


class RayDataExecutor(BaseExecutor):
    """Ray Data-based executor for pipeline execution.

    This executor:
    1. Executes setup on all nodes for all stages
    2. Converts initial tasks to Ray Data dataset
    3. Applies each stage as a Ray Data transformation (as a task or actor in map_batches)
    4. Returns final results as a list of tasks
    """

    def __init__(self, config: dict[str, Any] | None = None):
        super().__init__(config)
        logger.warning("Ray Data executor is experimental and might not work as expected.")

    def execute(self, stages: list["ProcessingStage"], initial_tasks: list[Task] | None = None) -> list[Task]:
        """Execute the pipeline stages using Ray Data.

        Args:
            stages (list[ProcessingStage]): List of processing stages to execute
            initial_tasks (list[Task], optional): Initial tasks to process (can be None for empty start)

        Returns:
            list[Task]: List of final processed tasks
        """
        if not stages:
            return []

        register_loguru_serializer()
        # This prevents verbose logging from Ray Data about serialization of the dataclass
        DataContext.get_current().enable_fallback_to_arrow_object_ext_type = True
        # Initialize with initial tasks if provided, otherwise start with EmptyTask
        tasks: list[Task] = initial_tasks if initial_tasks else [EmptyTask]
        output_tasks: list[Task] = []
        try:
            # Initialize ray and explicitly set NOSET to empty
            # This ensures if Xenna was used before which was setting NOSET, we end up overriding it.
            ray.init(
                ignore_reinit_error=True, runtime_env={"env_vars": {"RAY_EXPERIMENTAL_NOSET_CUDA_VISIBLE_DEVICES": ""}}
            )

            # Convert tasks to dataset
            current_dataset = self._tasks_to_dataset(tasks)

            # Execute setup on node for all stages
            execute_setup_on_node(stages)
            logger.info(f"Setup on node complete for all stages. Starting Ray Data pipeline with {len(stages)} stages")

            # Process through each stage
            for i, stage in enumerate(stages):
                # TODO: add pipeline level config for verbosity
                logger.info(f"Processing stage {i + 1}/{len(stages)}: {stage}")
                logger.info(f"  CPU cores: {stage.resources.cpus}, GPU ratio: {stage.resources.gpus}")

                # Create adapter for this stage
                adapter = RayDataStageAdapter(stage)

                # Apply stage transformation
                current_dataset = adapter.process_dataset(current_dataset)
        except Exception as e:
            logger.error(f"Error during pipeline execution: {e}")
            raise
        else:
            # Convert final dataset back to tasks
            # TODO: add pipeline configuration to check if user wants to return last stages output to driver
            output_tasks = self._dataset_to_tasks(current_dataset)
            logger.info(f"Pipeline completed. Final results: {len(output_tasks)} tasks")
        finally:
            # This ensures we unset all the env vars set above during initialize and kill the pending actors.
            ray.shutdown()
        return output_tasks

    def _tasks_to_dataset(self, tasks: list[Task]) -> Dataset:
        """Convert list of tasks to Ray Data dataset.

        Args:
            tasks: List of Task objects

        Returns:
            Ray Data dataset containing Task objects directly
        """
        # Create Ray Data dataset directly from Task objects
        return ray.data.from_items(tasks, override_num_blocks=len(tasks))

    def _dataset_to_tasks(self, dataset: Dataset) -> list[Task]:
        """Convert Ray Data dataset back to list of tasks.

        Args:
            dataset: Ray Data dataset containing Task objects

        Returns:
            List of Task objects
        """
        # Get all items from dataset
        items = dataset.take_all()

        # Handle the fact that Ray Data might return different formats
        tasks = []
        for item in items:
            tasks.append(item["item"])
        return tasks



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Literal

import numpy as np
from loguru import logger

from nemo_curator.backends.base import NodeInfo, WorkerMetadata
from nemo_curator.models.clip import CLIPAestheticScorer
from nemo_curator.stages.base import ProcessingStage
from nemo_curator.stages.resources import Resources
from nemo_curator.tasks.video import VideoTask
from nemo_curator.utils.decoder_utils import FrameExtractionPolicy, FrameExtractionSignature


@dataclass
class ClipAestheticFilterStage(ProcessingStage[VideoTask, VideoTask]):
    """Stage for filtering video clips based on CLIP aesthetic score.

    This class processes video clips through a series of steps including aesthetic score
    calculation and filtering based on thresholds.
    """

    model_dir: str = "models/clip_aesthetic"
    score_threshold: float = 0.5
    reduction: Literal["mean", "min"] = "min"
    target_fps: float = 1.0
    num_gpus_per_worker: float = 0.25
    verbose: bool = False
    _name: str = "clip_aesthetic_filter"

    def inputs(self) -> tuple[list[str], list[str]]:
        return ["data"], ["clips"]

    def outputs(self) -> tuple[list[str], list[str]]:
        return ["data"], ["decoded_motion_data"]

    def __post_init__(self) -> None:
        self._resources = Resources(gpus=self.num_gpus_per_worker)

    def setup_on_node(self, node_info: NodeInfo, worker_metadata: WorkerMetadata) -> None:  # noqa: ARG002
        """Download the weights for the CLIPAestheticScorer model on the node."""
        CLIPAestheticScorer.download_weights_on_node(self.model_dir)

    def setup(self, worker_metadata: WorkerMetadata | None = None) -> None:  # noqa: ARG002
        self.model = CLIPAestheticScorer(model_dir=self.model_dir)
        self.model.setup()
        self.frame_extraction_signature = FrameExtractionSignature(
            extraction_policy=FrameExtractionPolicy.sequence,
            target_fps=self.target_fps,
        ).to_str()
        if self.reduction == "mean":
            self.reduction_fn = np.mean
        elif self.reduction == "min":
            self.reduction_fn = np.min
        else:
            msg = f"Invalid reduction: {self.reduction}"
            raise ValueError(msg)

    def process(self, task: VideoTask) -> VideoTask:
        video = task.data
        passed_clips = []
        for clip in video.clips:
            if not clip.buffer:
                logger.warning(f"Clip {clip.uuid} has no buffer.")
                clip.errors["buffer"] = "empty"
                clip.aesthetic_score = -1.0
            elif self.frame_extraction_signature not in clip.extracted_frames:
                clip.errors[f"frames-{self.frame_extraction_signature}"] = "missing"
                error_msg = (
                    f"Clip {clip.uuid} has buffer but no extracted frames for {self.frame_extraction_signature}"
                )
                logger.error(error_msg)
                clip.aesthetic_score = -1.0
            else:
                frames = clip.extracted_frames.pop(self.frame_extraction_signature)
                scores = self.model(frames).cpu().numpy()
                del frames
                clip.aesthetic_score = float(self.reduction_fn(scores))

            # Filtering
            if clip.aesthetic_score < self.score_threshold:
                video.filtered_clips.append(clip)
                video.clip_stats.num_filtered_by_aesthetic += 1
                if self.verbose:
                    logger.info(
                        f"Clip {clip.uuid} has aesthetic score {clip.aesthetic_score:.3f} below threshold "
                        f"{self.score_threshold}, skipped.",
                    )
            else:
                passed_clips.append(clip)
                if self.verbose:
                    logger.info(
                        f"Clip {clip.uuid} has aesthetic score {clip.aesthetic_score:.3f} above threshold "
                        f"{self.score_threshold}, kept.",
                    )

        video.clips = passed_clips
        if self.verbose:
            logger.info(
                f"Video {video.input_video} chunk-{video.clip_chunk_index} has "
                f"{len(video.clips)}/{len(video.filtered_clips)} clips "
                "passed/filtered",
            )

        # @aot TODO: free memory periodically
        return task



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
from dataclasses import dataclass
from typing import cast

import av
import cv2
import numpy.typing as npt
import torch
from numpy.lib.recfunctions import structured_to_unstructured
from torch import Tensor

# We error on any video with a width or height less than this.
# The motion detection algorithm can't handle any resolutions less than this.
_MIN_SIDE_RESOLUTION = 256


class VideoResolutionTooSmallError(Exception):
    """Exception raised when video resolution is below the minimum required size.

    This error occurs when either the width or height of the video is less than
    the minimum resolution threshold required for motion detection.
    """


@dataclass
class MotionInfo:
    """Container for motion detection results.

    This class stores the results of motion detection analysis, including:
    - Whether the video has small motion
    - The minimum motion value in a 256x256 patch
    - The global average motion value across the entire videoq
    """

    is_small_motion: bool
    per_patch_min_256: float
    global_mean: float


@dataclass
class DecodedData:
    """Container for decoded video frames containing motion vector data.

    This class stores a list of decoded frames, each containing motion vector data,
    and the dimensions of the RGB decoded frame used to construct the flow vector.
    """

    # List of decoded frames containing motion vector data, not RGB
    frames: list[npt.NDArray]  # type: ignore[type-arg]
    # pass the dimensions of the RGB decoded frame to construct flow vector
    frame_size: torch.Size

    def get_major_size(self) -> int:
        """Calculate total size in bytes of all frames in the decoded data.

        Returns:
            Total size in bytes.

        """
        total_size = 0
        for frame in self.frames:
            total_size += frame.nbytes
        return total_size


def motion_vectors_to_flowfield(mvs: Tensor, size: list[int], flow: Tensor | None = None) -> Tensor:
    """Compute a canonical flow from motion vectors."""
    # get relevant info for later
    batch_size = mvs.shape[0]
    n_vectors = mvs.shape[1]
    block_sizes = mvs[..., 0:2]
    dst = mvs[..., 4:6]
    motion = mvs[..., 7:9]
    motion_scale = mvs[..., 9].unsqueeze(-1)
    device = mvs.device

    # Make indices for the batch number, this will be used as part of an index later
    batch_indices = torch.arange(batch_size, device=device)
    batch_pos = batch_indices.view(-1, 1).expand(-1, n_vectors).view(batch_size, n_vectors, 1)

    # compute sub-pixel src locations
    # `src = dst + motion / motion_scale`
    delta = -motion / motion_scale

    # add batch position to the source x, y positions
    dst_complete = torch.cat([batch_pos, dst], dim=-1)

    # create a "flow" where each source pixel is annotated with its destination pixel
    if flow is None or flow.shape != (batch_size, *size, 2):
        flow = torch.zeros(batch_size, *size, 2, device=device)
    else:
        flow.zero_()

    # choices for block size
    block_options = [
        torch.as_tensor([8, 8], device=device),
        torch.as_tensor([16, 16], device=device),
        torch.as_tensor([16, 8], device=device),
        torch.as_tensor([8, 16], device=device),
    ]

    # separate motion vectors by block size
    selected_blocks = [(block_sizes == b).all(-1).unsqueeze(-1) for b in block_options]

    # make offsets from center point for each of the block sizes.
    # this is ugly because we need to prepend 0 as a "batch offset" since the points are (batch, x, y)
    offsets = [
        torch.stack(torch.meshgrid(*[torch.arange(-b // 2, b // 2, device=device) for b in bs]))
        for bs in block_options
    ]

    # tile each of the mvs so that they contain the center point repeated for the size of the entire block
    dst_tiles = [
        dst_complete.masked_select(blocks).view(-1, 1, 1, 1).tile(1, 1, *block_size).view(-1, 3, *block_size)
        for blocks, block_size in zip(selected_blocks, block_options, strict=False)
    ]

    # # split each tensor into three indices, batch, h, w
    dst_b, dst_x, dst_y = zip(*[d.split(1, dim=1) for d in dst_tiles], strict=False)  # type: ignore[no-untyped-call]

    # Add in the offsets
    dst_x1 = [(mv_b + offset_b[0]) for mv_b, offset_b in zip(dst_x, offsets, strict=False)]
    dst_y1 = [(mv_b + offset_b[1]) for mv_b, offset_b in zip(dst_y, offsets, strict=False)]

    # check bounds (needed for indexing)
    dst_x2 = [mv_b.where(mv_b > 0, torch.as_tensor(0, device=device)) for mv_b in dst_x1]
    dst_y2 = [mv_b.where(mv_b > 0, torch.as_tensor(0, device=device)) for mv_b in dst_y1]

    dst_x3 = [mv_b.where(mv_b < size[-1], size[-1] - 1) for mv_b in dst_x2]
    dst_y3 = [mv_b.where(mv_b < size[-2], size[-2] - 1) for mv_b in dst_y2]

    # flatten the indices and concat them
    dst_x4 = torch.cat([mv.flatten() for mv in dst_x3], dim=0)
    dst_y4 = torch.cat([mv.flatten() for mv in dst_y3], dim=0)
    dst_b4 = torch.cat([mv.flatten() for mv in dst_b], dim=0)

    # tile and offset the source positions
    delta_tiles = [
        delta.masked_select(blocks).view(-1, 1, 1).tile(1, *block_size).view(-1, 2, *block_size)
        for blocks, block_size in zip(selected_blocks, block_options, strict=False)
    ]
    delta_flat = torch.cat([mv.movedim(1, -1).flatten(0, 2) for mv in delta_tiles], dim=0)

    # index the flow image and set it to the destination
    flow.index_put_((dst_b4.long(), dst_y4.long(), dst_x4.long()), delta_flat, accumulate=False)

    return flow


def decode_for_motion(  # noqa: C901
    video: io.BytesIO,
    thread_count: int = 4,
    target_fps: float = 2.0,
    target_duration_ratio: float = 0.5,
) -> DecodedData:
    """Decode video for motion detection.

    This function decodes a video stream to extract motion vectors.

    Args:
        video: Input video stream as a bytes object.
        thread_count: Number of threads to use for decoding.
        target_fps: Target frames per second for motion detection.
        target_duration_ratio: Ratio of target duration to source duration.

    Returns:
        DecodedData object containing motion vectors and frame dimensions.

    """
    with cast("av.container.InputContainer", av.open(video, metadata_errors="ignore")) as input_container:
        stream = input_container.streams.video[0]
        ctx = stream.codec_context
        # Set this flag to return motion vectors
        ctx.flags2 |= av.codec.context.Flags2.EXPORT_MVS
        ctx.thread_type = av.codec.context.ThreadType.AUTO
        ctx.thread_count = thread_count
        mv_data = []
        shape = torch.Size([1, 1])

        # Get video framerate and duration
        if stream.average_rate:
            source_fps = float(stream.average_rate)
        else:
            # Use nominal base rate if average not available
            source_fps = float(stream.base_rate if stream.base_rate else 30)

        # Get duration in seconds
        duration_seconds = 0
        if stream.duration is not None and stream.time_base is not None:
            duration_seconds = stream.duration * int(stream.time_base)
        if duration_seconds <= 0 and hasattr(input_container, "duration") and input_container.duration is not None:
            duration_seconds = int(input_container.duration / int(av.time_base))
        if duration_seconds <= 0:
            # Default assumption if we can't determine duration
            duration_seconds = 30

        # Calculate how many frames to sample based on ratio
        target_seconds = duration_seconds * target_duration_ratio
        max_frames = max(10, round(target_fps * target_seconds))  # At least 10 frames regardless of ratio

        # Calculate frame step to achieve target FPS
        # If source is 30fps and target is 2fps, we sample every 15 frames
        sample_step = max(1, round(source_fps / target_fps))

        for i, frame in enumerate(input_container.decode(video=0)):
            # Only process frames at our target step interval
            if i % sample_step != 0:
                continue

            # Process shape info from first frame
            if i == 0:
                shape = torch.Size([frame.height, frame.width, 3])
                if shape[0] < _MIN_SIDE_RESOLUTION or shape[1] < _MIN_SIDE_RESOLUTION:
                    error_msg = f"Expected min resolution of {_MIN_SIDE_RESOLUTION}, but got a resolution of {shape}"
                    raise VideoResolutionTooSmallError(error_msg)

            # Process motion vectors
            for sd in frame.side_data:
                if sd.type == av.sidedata.sidedata.Type.MOTION_VECTORS:  # type: ignore[attr-defined]
                    mv = structured_to_unstructured(
                        sd.to_ndarray(),  # type: ignore[attr-defined]
                    )
                    mv_data.append(mv)
                    break  # Skip checking other side data once we find motion vectors

            # Early stopping optimization: limit based on calculated max frames
            if len(mv_data) >= max_frames:
                break

    return DecodedData(frames=mv_data, frame_size=shape)


def check_if_small_motion(  # noqa: PLR0913
    mv_list: list[npt.NDArray],  # type: ignore[type-arg]
    frame_shape: torch.Size,
    global_mean_threshold: float = 0.00098,
    per_patch_min_256_threshold: float = 0.000001,
    *,
    use_gpu: bool = False,
    batch_size: int = 256,
) -> MotionInfo:
    """Check if a video has small motion.

    This function checks if a video has small motion by calculating the global mean
    and per-pixel average motion values.

    Args:
        mv_list: List of motion vectors.
        frame_shape: Shape of the frame.
        global_mean_threshold: Threshold for global mean motion.
        per_patch_min_256_threshold: Threshold for per-patch minimum motion.
        use_gpu: Whether to use GPU for computation.
        batch_size: Size of the batch for processing.

    Returns:
        MotionInfo object containing the results of the motion detection.

    """
    device = torch.device("cuda" if use_gpu else "cpu")

    global_sum_tensor = torch.tensor(0.0, device=device)
    per_pixel_sum_tensor = torch.zeros((frame_shape[0], frame_shape[1]), device=device)
    num_frames = 0
    preallocated_flow = torch.zeros(batch_size, frame_shape[0], frame_shape[1], 2, device=device)

    for batch_offset in range(0, len(mv_list), batch_size):
        current_batch_size = min(batch_size, len(mv_list) - batch_offset)

        # Pad all tensors to the same dimensions (the number of blocks varies per frame).
        # Padding with zeros is valid because the blocks of size (0, 0) will not be selected.
        max_n_vectors = max(mv_item.shape[0] for mv_item in mv_list[batch_offset : batch_offset + current_batch_size])
        mv_data_padded = torch.zeros(current_batch_size, max_n_vectors, 10, dtype=torch.float32, device=device)

        for batch_id in range(batch_offset, batch_offset + current_batch_size):
            data = torch.tensor(mv_list[batch_id][:, 1:], dtype=torch.float32, device=device)
            mv_data_padded[batch_id - batch_offset, : data.shape[0], :] = data

        flow_field = motion_vectors_to_flowfield(mv_data_padded, [frame_shape[0], frame_shape[1]], preallocated_flow)
        magnitudes = torch.linalg.vector_norm(flow_field, dim=3) / (frame_shape[1] + frame_shape[0])

        global_sum_tensor += magnitudes.sum()
        per_pixel_sum_tensor += magnitudes.sum(dim=0)
        num_frames += current_batch_size

    total_elements = num_frames * frame_shape[0] * frame_shape[1]
    global_mean = (global_sum_tensor / total_elements).item()
    per_pixel_avg = (per_pixel_sum_tensor / num_frames).cpu().numpy()
    per_patch_min_256 = float(cv2.resize(per_pixel_avg, None, fx=1 / 256, fy=1 / 256).min())

    is_small_motion = global_mean < global_mean_threshold or per_patch_min_256 < per_patch_min_256_threshold

    return MotionInfo(is_small_motion, per_patch_min_256, global_mean)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field
from typing import Any

from loguru import logger

from .tasks import Task


@dataclass
class FileGroupTask(Task[list[str]]):
    """Task representing a group of files to be read.
    This is created during the planning phase and passed to reader stages.
    """

    reader_config: dict[str, Any] = field(default_factory=dict)
    data: list[str] = field(default_factory=list)

    @property
    def num_items(self) -> int:
        """Number of files in this group."""
        return len(self.data)

    def validate(self) -> bool:
        """Validate the task data."""
        # TODO: We should fsspec checks for that file paths do exist
        # Handle both Python lists and numpy arrays by checking length
        # instead of boolean evaluation (which is ambiguous for numpy arrays)
        if len(self.data) == 0:
            logger.warning(f"No files to process in task {self.task_id}")
            return False
        if not isinstance(self.data, list):
            err = f"Invalid data type in task {self.task_id}"
            raise TypeError(err)
        return True



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Literal

import cudf
import numpy as np
import rmm

from nemo_curator.stages.base import ProcessingStage
from nemo_curator.stages.deduplication.fuzzy.utils import CURATOR_DEFAULT_MINHASH_FIELD
from nemo_curator.stages.deduplication.id_generator import CURATOR_DEDUP_ID_STR, get_id_generator_actor
from nemo_curator.stages.deduplication.io_utils import DeduplicationIO
from nemo_curator.stages.resources import Resources
from nemo_curator.tasks import FileGroupTask
from nemo_curator.utils.file_utils import create_or_overwrite_dir, get_fs

if TYPE_CHECKING:
    from nemo_curator.backends.base import WorkerMetadata


class MinHash(ABC):
    """
    Base class for computing minhash signatures of a document corpus
    """

    def __init__(
        self,
        seed: int = 42,
        num_hashes: int = 260,
        char_ngrams: int = 24,
        use_64bit_hash: bool = False,
    ):
        """
        Parameters
        ----------
        seed: Seed for minhash permutations
        num_hashes: Length of minhash signature (No. of minhash permutations)
        char_ngrams: Width of text window (in characters) while computing minhashes.
        use_64bit_hash: Whether to use a 64 bit hash function.
        """
        self.num_hashes = num_hashes
        self.char_ngram = char_ngrams
        self.seed = seed
        self.use_64bit_hash = use_64bit_hash

    def generate_seeds(self, n_permutations: int = 260, seed: int = 0, bit_width: int = 32) -> np.ndarray:
        """
        Generate seeds for all minhash permutations based on the given seed.
        This is a placeholder that child classes should implement if needed.
        """
        msg = "Child classes should implement this method if needed"
        raise NotImplementedError(msg)

    @abstractmethod
    def compute_minhashes(self, text_series: Any) -> Any:  # noqa: ANN401
        """
        Compute minhash signatures for the given dataframe text column.
        """


class GPUMinHash(MinHash):
    def __init__(
        self,
        seed: int = 42,
        num_hashes: int = 260,
        char_ngrams: int = 24,
        use_64bit_hash: bool = False,
        pool: bool = False,
    ):
        # Initialize parent class
        MinHash.__init__(
            self,
            seed=seed,
            num_hashes=num_hashes,
            char_ngrams=char_ngrams,
            use_64bit_hash=use_64bit_hash,
        )

        # Initialize memory pool for cuDF
        if pool:
            rmm.reinitialize(pool_allocator=pool)

        # Generate seeds
        self.seeds = self.generate_seeds(
            n_permutations=self.num_hashes,
            seed=self.seed,
            bit_width=64 if self.use_64bit_hash else 32,
        )

    def generate_seeds(self, n_permutations: int = 260, seed: int = 0, bit_width: int = 32) -> np.ndarray:
        """
        Generate seeds for all minhash permutations based on the given seed.
        """
        gen = np.random.RandomState(seed)

        if bit_width == 32:  # noqa: PLR2004
            MERSENNE_PRIME = np.uint32((1 << 31) - 1)  # noqa: N806
            dtype = np.uint32
        elif bit_width == 64:  # noqa: PLR2004
            # For 64-bit, use a larger prime number suitable for 64-bit operations
            MERSENNE_PRIME = np.uint64((1 << 61) - 1)  # noqa: N806
            dtype = np.uint64
        else:
            msg = "Unsupported bit width. Use either 32 or 64."
            raise ValueError(msg)

        return np.array(
            [
                (
                    gen.randint(1, MERSENNE_PRIME, dtype=dtype),
                    gen.randint(0, MERSENNE_PRIME, dtype=dtype),
                )
                for _ in range(n_permutations)
            ],
            dtype=dtype,
        )

    def minhash32(self, ser: cudf.Series) -> cudf.Series:
        """
        Compute 32bit minhashes based on the MurmurHash3 algorithm
        """
        if not isinstance(ser, cudf.Series):
            msg = "Expected data of type cudf.Series"
            raise TypeError(msg)

        seeds_a = cudf.Series(self.seeds[:, 0], dtype="uint32")
        seeds_b = cudf.Series(self.seeds[:, 1], dtype="uint32")

        return ser.str.minhash(a=seeds_a, b=seeds_b, seed=self.seeds[0][0], width=self.char_ngram)

    def minhash64(self, ser: cudf.Series) -> cudf.Series:
        """
        Compute 64bit minhashes based on the MurmurHash3 algorithm
        """
        if not isinstance(ser, cudf.Series):
            msg = "Expected data of type cudf.Series"
            raise TypeError(msg)

        seeds_a = cudf.Series(self.seeds[:, 0], dtype="uint64")
        seeds_b = cudf.Series(self.seeds[:, 1], dtype="uint64")

        return ser.str.minhash64(a=seeds_a, b=seeds_b, seed=self.seeds[0][0], width=self.char_ngram)

    def compute_minhashes(self, text_series: cudf.Series) -> cudf.Series:
        """
        Compute minhash signatures for the given text series.

        Parameters
        ----------
        text_series: cudf.Series
            Series containing text data to compute minhashes for

        Returns
        -------
        cudf.Series containing minhash signatures
        """
        if not isinstance(text_series, cudf.Series):
            msg = "Expected data of type cudf.Series"
            raise TypeError(msg)

        # Compute minhashes
        minhash_method = self.minhash64 if self.use_64bit_hash else self.minhash32
        return minhash_method(text_series)


class MinHashStage(ProcessingStage[FileGroupTask, FileGroupTask], DeduplicationIO):
    """
    ProcessingStage for computing MinHash signatures on documents for fuzzy deduplication.

    This stage takes FileGroupTask containing paths to input documents and produces
    FileGroupTask containing paths to computed minhash signature files. It uses GPU-accelerated
    MinHash computation to generate locality-sensitive hash signatures that can be used
    for approximate duplicate detection.

    The stage automatically handles:
    - Reading input files (JSONL or Parquet format)
    - Assigning unique Integer IDs to documents using the IdGenerator actor
    - Computing MinHash signatures using GPU acceleration
    - Writing results to Parquet files

    Parameters
    ----------
    output_path : str
        Base path where minhash output files will be written
    text_field : str, default="text"
        Name of the field containing text to compute minhashes from
    minhash_field : str, default="_minhash_signature"
        Name of the field where minhash signatures will be stored
    char_ngrams : int, default=24
        Width of character n-grams for minhashing
    num_hashes : int, default=260
        Number of hash functions (length of minhash signature)
    seed : int, default=42
        Random seed for reproducible minhash generation
    use_64bit_hash : bool, default=False
        Whether to use 64-bit hash functions (vs 32-bit)
    read_format : Literal["jsonl", "parquet"], default="jsonl"
        Format of input files
    read_kwargs : dict[str, Any] | None, default=None
        Additional keyword arguments for reading input files
    write_kwargs : dict[str, Any] | None, default=None
        Additional keyword arguments for writing output files

    Examples
    --------
    >>> stage = MinHashStage(
    ...     output_path="/path/to/minhash/output",
    ...     text_field="content",
    ...     num_hashes=128,
    ...     char_ngrams=5
    ... )
    >>> # Use in a pipeline to process document batches
    """

    def __init__(  # noqa: PLR0913
        self,
        output_path: str,
        text_field: str = "text",
        minhash_field: str = CURATOR_DEFAULT_MINHASH_FIELD,
        char_ngrams: int = 24,
        num_hashes: int = 260,
        seed: int = 42,
        use_64bit_hash: bool = False,
        read_format: Literal["jsonl", "parquet"] = "jsonl",
        read_kwargs: dict[str, Any] | None = None,
        write_kwargs: dict[str, Any] | None = None,
        pool: bool = True,
    ):
        # Set ProcessingStage attributes
        self._name = self.__class__.__name__
        self._resources = Resources(gpus=1.0)  # Requires 1 GPU

        self.text_field = text_field
        self.minhash_field = minhash_field
        self.char_ngrams = char_ngrams
        self.num_hashes = num_hashes
        self.seed = seed
        self.use_64bit_hash = use_64bit_hash
        self.read_format = read_format
        self.read_kwargs = read_kwargs or {}
        self.write_kwargs = write_kwargs or {}
        self.pool = pool
        # Initialize the minhash processor in setup
        self.minhash_processor = None
        self.id_generator = None

        self.output_fs = get_fs(output_path, self.write_kwargs.get("storage_options", {}))
        self.output_path = self.output_fs.sep.join([output_path, self._name])
        create_or_overwrite_dir(self.output_path, storage_options=self.write_kwargs.get("storage_options", {}))

    def setup(self, _worker_metadata: "WorkerMetadata | None" = None) -> None:
        """Initialize the GPU MinHash processor and ID generator."""
        # Initialize the ID generator (will be shared across workers)

        try:
            self.id_generator = get_id_generator_actor()
        except ValueError as e:
            err_msg = """
            Failed to get ID generator actor. Start an ID generator actor via `create_id_generator_actor` if calling this stage directly.
            If using the FuzzyDedup API this should be started automatically.
            """
            raise ValueError(err_msg) from e

        # Initialize the GPU minhash processor
        self.minhash_processor = GPUMinHash(
            seed=self.seed,
            num_hashes=self.num_hashes,
            char_ngrams=self.char_ngrams,
            use_64bit_hash=self.use_64bit_hash,
            pool=self.pool,
        )

    def inputs(self) -> tuple[list[str], list[str]]:
        """Define input requirements."""
        return (["data"], [])

    def outputs(self) -> tuple[list[str], list[str]]:
        """Define outputs - produces FileGroupTask with minhash files."""
        return (["data"], [])

    def process(self, task: FileGroupTask) -> FileGroupTask:
        """
        Process a group of files to compute minhashes.

        Args:
            task: FileGroupTask containing file paths to process

        Returns:
            FileGroupTask containing paths to minhash output files
        """

        if self.minhash_processor is None or self.id_generator is None:
            msg = "MinHash processor or ID generator not initialized. Call setup() first."
            raise RuntimeError(msg)

        output_file = self.output_fs.sep.join([self.output_path, f"{task._uuid}.parquet"])

        read_kwargs = self.read_kwargs.copy()

        # Read input file based on format
        if self.read_format == "jsonl":
            df = self.read_jsonl(filepath=task.data, columns=[self.text_field], assign_id=True, **read_kwargs)
        elif self.read_format == "parquet":
            df = self.read_parquet(filepath=task.data, columns=[self.text_field], assign_id=True, **read_kwargs)
        else:
            msg = f"Unsupported read format: {self.read_format}"
            raise ValueError(msg)

        result_df = df[[CURATOR_DEDUP_ID_STR]]
        result_df[self.minhash_field] = self.minhash_processor.compute_minhashes(df[self.text_field])

        # Write output file
        self.write_parquet(df=result_df, filepath=output_file, **self.write_kwargs)

        # Return FileGroupTask with output file
        return FileGroupTask(
            task_id=f"{task.task_id}",
            dataset_name=f"{task.dataset_name}_minhash",
            data=[output_file],
            _metadata={
                **task._metadata,
                "minhash_field": self.minhash_field,
                "num_hashes": self.num_hashes,
                "storage_options": self.write_kwargs.get("storage_options"),
            },
            _stage_perf=task._stage_perf,
        )



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import ClassVar

import justext
import lxml
from loguru import logger

from .base import HTMLExtractorAlgorithm


class JusTextExtractor(HTMLExtractorAlgorithm):
    # Class-level set to track which languages we've already logged warnings for
    _logged_languages: ClassVar[set[str]] = set()

    def __init__(  # noqa: PLR0913
        self,
        length_low: int = 70,
        length_high: int = 200,
        stopwords_low: float = 0.30,
        stopwords_high: float = 0.32,
        max_link_density: float = 0.2,
        max_heading_distance: int = 200,
        no_headings: bool = False,
        is_boilerplate: bool | None = None,
    ):
        """
        Initialize the jusText text extraction algorithm with specified parameters.

        jusText is a tool for removing boilerplate content, such as navigation links, headers, and footers from HTML pages.
        It is designed to preserve mainly text containing full sentences and it is therefore well suited for creating linguistic resources such as Web corpora.
        The key idea is that long blocks can often be classified with high confidence, while shorter blocks require context-based adjustments.

        Here is an overview of the jusText algorithm:
            • Segmentation: The document is split into textual blocks based on HTML tags that typically define separate sections (e.g., <div>, <p>, <table>).
            • Preprocessing: Contents of <header>, <style>, and <script> tags are removed.
                Certain elements (e.g., <select>, copyright symbols) are immediately classified as boilerplate.
            • Context-Free Classification: Each block is classified as:
                - Bad (boilerplate) if it has high link density.
                - Short if it is too small to be classified reliably.
                - Near-Good if it has a moderate density of stopwords.
                - Good (main content) if it is long and contains many stopwords.
            • Context-Sensitive Classification: Blocks that were classified as short or near-good are reclassified based on surrounding blocks.
                The assumption is that main content clusters together, as does boilerplate.
            • Headings Processing: Header elements (e.g., <h1>, <h2>) are treated separately to ensure useful headings are preserved.
                Short headers near good content may be reclassified as near-good or good.

        Please refer to the jusText documentation for more details: https://corpus.tools/wiki/Justext/Algorithm

        Args:
            length_low: Minimum length of text to be considered for extraction.
            length_high: Maximum length of text to be considered for extraction.
            stopwords_low: Minimum proportion of stopwords in the text to be considered for extraction.
            stopwords_high: Maximum proportion of stopwords in the text to be considered for extraction.
            max_link_density: Maximum allowed link density in the text.
            max_heading_distance: Maximum distance from a heading to consider text for extraction.
            no_headings: If True, text extraction will ignore headings.
            is_boilerplate: If True, text extraction will ignore boilerplate content.
                Default is True for space-separated languages and False for non-space-separated languages
                (Thai, Chinese, Japanese, and Korean).
            logger: Optional logger instance for logging messages.

        """
        self.length_low = length_low
        self.length_high = length_high
        self.stopwords_low = stopwords_low
        self.stopwords_high = stopwords_high
        self.max_link_density = max_link_density
        self.max_heading_distance = max_heading_distance
        self.no_headings = no_headings
        self.is_boilerplate = is_boilerplate

    def extract_text(self, html: str, stop_words: frozenset[str], language: str) -> list[str] | None:
        # Segment the HTML into paragraphs
        try:
            # Form the DOM tree
            dom = justext.core.html_to_dom(html)
            cleaned_dom = justext.core.preprocessor(dom)
            # Get the paragraphs from the DOM
            handler = justext.core.ParagraphMaker()
            lxml.sax.saxify(cleaned_dom, handler)
        except (lxml.etree.ParserError, ValueError, Exception):
            # Return nothing when we cannot segment the document
            logger.info("Could not segment paragaphs in the document")
            return None

        paragraphs = handler.paragraphs

        # Context free classification
        justext.core.classify_paragraphs(
            paragraphs,
            stop_words,
            self.length_low,
            self.length_high,
            self.stopwords_low,
            self.stopwords_high,
            self.max_link_density,
            self.no_headings,
        )

        # Copy the context free class to the class_style
        # This handles the headings as described in the
        # documentation
        for paragraph in paragraphs:
            paragraph.class_type = paragraph.cf_class

        # Context sensitive classification
        justext.core.revise_paragraph_classification(
            paragraphs,
            self.max_heading_distance,
        )

        if self.is_boilerplate is None:
            if language in self.NON_SPACED_LANGUAGES:
                if language not in self._logged_languages:
                    logger.warning(f"Disabling is_boilerplate check for jusText extraction for language: {language}")
                    self._logged_languages.add(language)
                is_boilerplate = False
            else:
                is_boilerplate = True

        else:
            is_boilerplate = self.is_boilerplate

        if is_boilerplate:
            return [p.text for p in paragraphs if not p.is_boilerplate]

        else:
            return [p.text for p in paragraphs]



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from dataclasses import dataclass

os.environ["RAPIDS_NO_INITIALIZE"] = "1"
import numpy as np
import pandas as pd
import torch
from huggingface_hub import PyTorchModelHubMixin
from torch import nn
from transformers import AutoModel

from nemo_curator.stages.base import CompositeStage, ProcessingStage
from nemo_curator.stages.text.models.model import ModelStage
from nemo_curator.stages.text.models.tokenizer import TokenizerStage
from nemo_curator.stages.text.models.utils import ATTENTION_MASK_COLUMN, INPUT_ID_COLUMN, format_name_with_suffix
from nemo_curator.tasks import DocumentBatch

from .constants import DEBERTA_TOKENIZER_PADDING_SIDE

PROMPT_TASK_COMPLEXITY_MODEL_IDENTIFIER = "nvidia/prompt-task-and-complexity-classifier"
MAX_SEQ_LENGTH = 512
OUTPUT_COLUMNS = [
    "prompt_complexity_score",
    "task_type_1",
    "task_type_2",
    "task_type_prob",
    "creativity_scope",
    "reasoning",
    "contextual_knowledge",
    "number_of_few_shots",
    "domain_knowledge",
    "no_label_reason",
    "constraint_ct",
]


class MeanPooling(nn.Module):
    def __init__(self):
        super().__init__()

    @torch.no_grad()
    def forward(self, last_hidden_state: torch.Tensor, attention_mask: torch.Tensor) -> torch.Tensor:
        input_mask_expanded = attention_mask.unsqueeze(-1).expand(last_hidden_state.size()).float()
        sum_embeddings = torch.sum(last_hidden_state * input_mask_expanded, 1)

        sum_mask = input_mask_expanded.sum(1)
        sum_mask = torch.clamp(sum_mask, min=1e-9)

        return sum_embeddings / sum_mask


class MulticlassHead(nn.Module):
    def __init__(self, input_size: int, num_classes: int) -> None:
        super().__init__()
        self.fc = nn.Linear(input_size, num_classes)

    @torch.no_grad()
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.fc(x)


class CustomDeberta(nn.Module, PyTorchModelHubMixin):
    def __init__(self, config: dataclass):
        super().__init__()

        self.backbone = AutoModel.from_pretrained(config["base_model"])
        self.target_sizes = config["target_sizes"].values()

        self.task_type_map = config["task_type_map"]
        self.weights_map = config["weights_map"]
        self.divisor_map = config["divisor_map"]

        self.heads = [MulticlassHead(self.backbone.config.hidden_size, sz) for sz in self.target_sizes]

        for i, head in enumerate(self.heads):
            self.add_module(f"head_{i}", head)

        self.pool = MeanPooling()

    @property
    def device(self) -> torch.device:
        return next(self.parameters()).device

    def compute_results(
        self, preds: torch.Tensor, target: str, decimal: int = 4
    ) -> tuple[list[str], list[str], list[float]]:
        if target == "task_type":
            top2_indices = torch.topk(preds, k=2, dim=1).indices
            softmax_probs = torch.softmax(preds, dim=1)
            top2_probs = softmax_probs.gather(1, top2_indices)
            top2 = top2_indices.detach().cpu().tolist()
            top2_prob = top2_probs.detach().cpu().tolist()

            top2_strings = [[self.task_type_map[str(idx)] for idx in sample] for sample in top2]
            top2_prob_rounded = [[round(value, 3) for value in sublist] for sublist in top2_prob]

            for counter, sublist in enumerate(top2_prob_rounded):
                if sublist[1] < 0.1:  # noqa: PLR2004
                    top2_strings[counter][1] = "NA"

            task_type_1 = [sublist[0] for sublist in top2_strings]
            task_type_2 = [sublist[1] for sublist in top2_strings]
            task_type_prob = [sublist[0] for sublist in top2_prob_rounded]

            return (task_type_1, task_type_2, task_type_prob)

        else:
            preds = torch.softmax(preds, dim=1)

            weights = np.array(self.weights_map[target])
            weighted_sum = np.sum(np.array(preds.detach().cpu()) * weights, axis=1)
            scores = weighted_sum / self.divisor_map[target]

            scores = [round(value, decimal) for value in scores]
            if target == OUTPUT_COLUMNS[7]:
                scores = [x if x >= 0.05 else 0 for x in scores]  # noqa: PLR2004
            return scores

    def process_logits(self, logits: list[torch.Tensor]) -> dict[str, torch.Tensor]:
        result = {}

        # Round 1: "task_type"
        task_type_logits = logits[0]
        task_type_results = self.compute_results(task_type_logits, target="task_type")
        result[OUTPUT_COLUMNS[1]] = task_type_results[0]
        result[OUTPUT_COLUMNS[2]] = task_type_results[1]
        result[OUTPUT_COLUMNS[3]] = task_type_results[2]

        # Round 2: "creativity_scope"
        creativity_scope_logits = logits[1]
        result[OUTPUT_COLUMNS[4]] = self.compute_results(creativity_scope_logits, target=OUTPUT_COLUMNS[4])

        # Round 3: "reasoning"
        reasoning_logits = logits[2]
        result[OUTPUT_COLUMNS[5]] = self.compute_results(reasoning_logits, target=OUTPUT_COLUMNS[5])

        # Round 4: "contextual_knowledge"
        contextual_knowledge_logits = logits[3]
        result[OUTPUT_COLUMNS[6]] = self.compute_results(contextual_knowledge_logits, target=OUTPUT_COLUMNS[6])

        # Round 5: "number_of_few_shots"
        number_of_few_shots_logits = logits[4]
        result[OUTPUT_COLUMNS[7]] = self.compute_results(number_of_few_shots_logits, target=OUTPUT_COLUMNS[7])

        # Round 6: "domain_knowledge"
        domain_knowledge_logits = logits[5]
        result[OUTPUT_COLUMNS[8]] = self.compute_results(domain_knowledge_logits, target=OUTPUT_COLUMNS[8])

        # Round 7: "no_label_reason"
        no_label_reason_logits = logits[6]
        result[OUTPUT_COLUMNS[9]] = self.compute_results(no_label_reason_logits, target=OUTPUT_COLUMNS[9])

        # Round 8: "constraint_ct"
        constraint_ct_logits = logits[7]
        result[OUTPUT_COLUMNS[10]] = self.compute_results(constraint_ct_logits, target=OUTPUT_COLUMNS[10])

        # Round 9: "prompt_complexity_score"
        result[OUTPUT_COLUMNS[0]] = torch.tensor(
            [
                round(
                    0.35 * creativity
                    + 0.25 * reasoning
                    + 0.15 * constraint
                    + 0.15 * domain_knowledge
                    + 0.05 * contextual_knowledge
                    + 0.05 * few_shots,
                    5,
                )
                for creativity, reasoning, constraint, domain_knowledge, contextual_knowledge, few_shots in zip(
                    result[OUTPUT_COLUMNS[4]],
                    result[OUTPUT_COLUMNS[5]],
                    result[OUTPUT_COLUMNS[10]],
                    result[OUTPUT_COLUMNS[8]],
                    result[OUTPUT_COLUMNS[6]],
                    result[OUTPUT_COLUMNS[7]],
                    strict=False,
                )
            ],
        )

        return result

    @torch.no_grad()
    def _forward(self, input_ids: torch.Tensor, attention_mask: torch.Tensor) -> dict[str, torch.Tensor]:
        outputs = self.backbone(input_ids=input_ids, attention_mask=attention_mask)

        last_hidden_state = outputs.last_hidden_state
        mean_pooled_representation = self.pool(last_hidden_state, attention_mask)

        logits = [self.heads[k](mean_pooled_representation) for k in range(len(self.target_sizes))]

        return self.process_logits(logits)

    @torch.no_grad()
    def forward(self, batch: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        input_ids = batch[INPUT_ID_COLUMN]
        attention_mask = batch[ATTENTION_MASK_COLUMN]

        return self._forward(input_ids, attention_mask)


class PromptTaskComplexityModelStage(ModelStage):
    """
    Stage for Hugging Face model inference.

    Args:
        cache_dir: The Hugging Face cache directory. Defaults to None.
        model_inference_batch_size: The size of the batch for model inference. Defaults to 256.
        has_seq_order: Whether to sort the input data by the length of the input tokens.
            Sorting is encouraged to improve the performance of the inference model. Defaults to True.
        autocast: Whether to use autocast. When True, we trade off minor accuracy for faster inference.
            Defaults to True.

    """

    def __init__(
        self,
        cache_dir: str | None = None,
        model_inference_batch_size: int = 256,
        has_seq_order: bool = True,
        autocast: bool = True,
    ):
        super().__init__(
            model_identifier=PROMPT_TASK_COMPLEXITY_MODEL_IDENTIFIER,
            cache_dir=cache_dir,
            has_seq_order=has_seq_order,
            model_inference_batch_size=model_inference_batch_size,
            padding_side=DEBERTA_TOKENIZER_PADDING_SIDE,
            unpack_inference_batch=False,
        )

        self.autocast = autocast

    def outputs(self) -> tuple[list[str], list[str]]:
        return ["data"], OUTPUT_COLUMNS

    def _setup(self, local_files_only: bool = True) -> None:
        self.model = (
            CustomDeberta.from_pretrained(
                self.model_identifier,
                cache_dir=self.cache_dir,
                local_files_only=local_files_only,
            )
            .cuda()
            .eval()
        )

    def process_model_output(self, outputs: torch.Tensor, _: dict[str, torch.Tensor] | None = None) -> torch.Tensor:
        return outputs

    def create_output_dataframe(self, df_cpu: pd.DataFrame, collected_output: dict[str, np.ndarray]) -> pd.DataFrame:
        df_cpu = df_cpu.drop(columns=[INPUT_ID_COLUMN, ATTENTION_MASK_COLUMN])

        for column in OUTPUT_COLUMNS:
            df_cpu[column] = collected_output[column]

        return df_cpu


@dataclass
class PromptTaskComplexityClassifier(CompositeStage[DocumentBatch, DocumentBatch]):
    """
    PromptTaskComplexityClassifier is a multi-headed model which classifies English text prompts across task types and complexity dimensions.
    Tasks are classified across 11 common categories. Complexity is evaluated across 6 dimensions and ensembled to create an overall complexity score.
    Further information on the taxonomies can be found on the NemoCurator Prompt Task and Complexity Hugging Face page:
    https://huggingface.co/nvidia/prompt-task-and-complexity-classifier.
    This class is optimized for running on multi-node, multi-GPU setups to enable fast and efficient inference on large datasets.

    Args:
        cache_dir: The Hugging Face cache directory. Defaults to None.
        text_field: The name of the text field in the input data. Defaults to "text".
        filter_by: For categorical classifiers, the list of labels to filter the data by. Defaults to None.
            Not supported with PromptTaskComplexityClassifier (raises NotImplementedError).
        max_chars: Limits the total number of characters that can be fed to the tokenizer.
            If None, text will not be truncated. Defaults to 2000.
        sort_by_length: Whether to sort the input data by the length of the input tokens.
            Sorting is encouraged to improve the performance of the inference model. Defaults to True.
        model_inference_batch_size: The size of the batch for model inference. Defaults to 256.
        autocast: Whether to use autocast. When True, we trade off minor accuracy for faster inference.
            Defaults to True.

    """

    cache_dir: str | None = None
    text_field: str = "text"
    filter_by: list[str] | None = None
    max_chars: int = 2000
    sort_by_length: bool = True
    model_inference_batch_size: int = 256
    autocast: bool = True

    def __post_init__(self) -> None:
        super().__init__()

        self._name = format_name_with_suffix(PROMPT_TASK_COMPLEXITY_MODEL_IDENTIFIER)

        if self.filter_by is not None and len(self.filter_by) > 0:
            msg = "filter_by not supported with PromptTaskComplexityClassifier"
            raise NotImplementedError(msg)

        self.stages = [
            TokenizerStage(
                model_identifier=PROMPT_TASK_COMPLEXITY_MODEL_IDENTIFIER,
                cache_dir=self.cache_dir,
                text_field=self.text_field,
                max_chars=self.max_chars,
                max_seq_length=MAX_SEQ_LENGTH,
                padding_side=DEBERTA_TOKENIZER_PADDING_SIDE,
                sort_by_length=self.sort_by_length,
            ),
            PromptTaskComplexityModelStage(
                cache_dir=self.cache_dir,
                model_inference_batch_size=self.model_inference_batch_size,
                has_seq_order=self.sort_by_length,
                autocast=self.autocast,
            ),
        ]

    def inputs(self) -> tuple[list[str], list[str]]:
        return self.stages[0].inputs()

    def outputs(self) -> tuple[list[str], list[str]]:
        return self.stages[1].outputs()

    def decompose(self) -> list[ProcessingStage]:
        return self.stages



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import cudf
    import cupy as cp

from typing import Any

import pyarrow.parquet as pq
from fsspec.parquet import open_parquet_file
from loguru import logger


def get_array_from_df(df: "cudf.DataFrame", embedding_col: str) -> "cp.ndarray":
    """
    Convert a column of lists to a 2D array.
    """
    return df[embedding_col].list.leaves.values.reshape(len(df), -1)


def break_parquet_partition_into_groups(
    files: list[str], embedding_dim: int | None = None, storage_options: dict[str, Any] | None = None
) -> list[list[str]]:
    """Break parquet files into groups to avoid cudf 2bn row limit."""
    if embedding_dim is None:
        # Default aggressive assumption of 1024 dimensional embedding
        embedding_dim = 1024

    cudf_max_num_rows = 2_000_000_000  # cudf only allows 2bn rows
    cudf_max_num_elements = cudf_max_num_rows / embedding_dim  # cudf considers each element in an array to be a row

    # Load the first file and get the number of rows to estimate
    with open_parquet_file(files[0], storage_options=storage_options) as f:
        # Multiply by 1.5 to adjust for skew
        avg_num_rows = pq.read_metadata(f).num_rows * 1.5

    max_files_per_subgroup = int(cudf_max_num_elements / avg_num_rows)
    max_files_per_subgroup = max(1, max_files_per_subgroup)  # Ensure at least 1 file per subgroup

    # Break files into subgroups
    subgroups = [files[i : i + max_files_per_subgroup] for i in range(0, len(files), max_files_per_subgroup)]
    if len(subgroups) > 1:
        logger.debug(
            f"Broke {len(files)} files into {len(subgroups)} subgroups with max {max_files_per_subgroup} files per subgroup"
        )
    return subgroups



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import subprocess
from dataclasses import dataclass

import torch
from torchvision import transforms  # type: ignore[import-untyped]
from torchvision.transforms import InterpolationMode, v2  # type: ignore[import-untyped]

from nemo_curator.utils.decoder_utils import decode_video_cpu_frame_ids, get_avg_frame_rate, get_frame_count
from nemo_curator.utils.operation_utils import make_pipeline_named_temporary_file


@dataclass
class WindowFrameInfo:
    """Container for frame window information, storing start and end frame indices.

    This class represents a window of frames in a video, defined by its start and end frame positions.
    """

    start: int
    end: int


WINDOW_MIN_FRAMES = 4


def compute_windows(total_frames: int, window_size: int = 128, remainder_threshold: int = 64) -> list[WindowFrameInfo]:
    """Generate windows by splitting the video into segments of the specified size.

    Args:
        total_frames: total frames
        window_size: The size of each window in number of frames.
        remainder_threshold: The minimum number of frames required to create a new window from the remainder.

    Yields:
        Tuple of (start_frame, end_frame) representing each window.

    """
    if not total_frames or total_frames < WINDOW_MIN_FRAMES:
        return []
    if total_frames <= window_size:
        return [WindowFrameInfo(0, total_frames - 1)]
    # Calculate the number of full window_size windows
    num_full_windows = total_frames // window_size

    # Calculate the remainder frames after filling in window_size windows
    remainder = total_frames % window_size

    out: list[WindowFrameInfo] = []
    # Yield each full window
    for i in range(num_full_windows):
        start_frame = i * window_size
        end_frame = start_frame + window_size - 1
        out.append(WindowFrameInfo(start_frame, end_frame))

    # Handle the remainder
    if remainder >= remainder_threshold:
        out.append(WindowFrameInfo(total_frames - remainder, total_frames - 1))
    elif remainder > 0 and num_full_windows > 0:
        # Expand the last window with the remainder if it exists
        out[-1] = WindowFrameInfo(out[-1].start, total_frames - 1)
    return out


def split_video_into_windows(  # noqa: PLR0913
    mp4_bytes: bytes,
    window_size: int = 256,
    remainder_threshold: int = 128,
    sampling_fps: float = 2.0,
    *,
    model_does_preprocess: bool = False,
    preprocess_dtype: str = "uint8",
    flip_input: bool = False,
    num_frames_to_use: int = 0,
    return_bytes: bool = False,
    return_video_frames: bool = True,
    num_threads: int = 1,
) -> tuple[list[bytes], list[torch.Tensor | None], list[WindowFrameInfo]]:
    """Calculate windows and return video inputs for language model from input clips.

    Processes video to determine the windows for a clip, decode in one shot and return processed frames
    for each window in a format suitable for consumption by the Qwen model.

    Args:
        mp4_bytes: input video in bytes
        fps: Frames per second of the input video.
        preprocess_dtype: Data type to use for preprocessing the video/image inputs.
        num_frames_to_use: Number of frames to extract from the video. If 0, uses all frames.
        flip_input: Whether to flip the input video/image horizontally.
        return_bytes: Whether to extract mp4 bytes for each window for use by PreviewStage
        model_does_preprocess: if the model does preprocessing
        num_threads: number of threads
        remainder_threshold: threshold for remainder
        return_video_frames: whether to return video frames
        sampling_fps: sampling fps
        window_size: window size

    Returns:
        Tuple containing:
            - "window_mp4_bytes": mp4 bytes corresponding to each window - only used when Preview stage is enabled
            - "window_frames": Decoded and per-window processed frames ready for use by Qwen model
            - "window info": start and end frame indices for each window in a clip

    """
    with make_pipeline_named_temporary_file(sub_dir="windowing") as input_file:
        input_file.write_bytes(mp4_bytes)
        total_frames = get_frame_count(mp4_bytes)
        windows = compute_windows(total_frames, window_size, remainder_threshold)
        video_frames: list[torch.Tensor | None] = []
        mp4_bytes_list: list[bytes] = []

        if not windows:
            return mp4_bytes_list, video_frames, windows

        if return_video_frames:
            video, frame_counts = fetch_video(
                str(input_file),
                sampling_fps=sampling_fps,
                window_range=windows,
                do_preprocess=not model_does_preprocess,
                preprocess_dtype=preprocess_dtype,
                num_frames_to_use=num_frames_to_use,
                flip_input=flip_input,
            )

            index = 0
            for count in frame_counts:
                video_frames.append(video[index : index + count - 1])
                index = count

        if return_bytes:
            if len(windows) == 1:
                return [mp4_bytes], video_frames, windows

            for window in windows:
                with make_pipeline_named_temporary_file(sub_dir="windowing") as tmp_file:
                    # Use ffmpeg to split the file on the frames.
                    command = [
                        "ffmpeg",
                        "-threads",
                        str(num_threads),
                        "-y",
                        "-i",
                        str(input_file),
                        "-loglevel",
                        "error",
                        "-vf",
                        f"select='between(n\\,{window.start}\\,{window.end})',setpts=PTS-STARTPTS",
                        "-threads",
                        str(num_threads),
                        "-f",
                        "mp4",
                        "-an",
                        str(tmp_file),
                    ]
                    subprocess.check_call(command)  # noqa: S603
                    mp4_bytes_list.append(tmp_file.read_bytes())
        return mp4_bytes_list, video_frames, windows


IMAGE_FACTOR = 28
MIN_PIXELS = 4 * 28 * 28
MAX_PIXELS = 16384 * 28 * 28
MAX_RATIO = 200

VIDEO_MIN_PIXELS = 128 * 28 * 28
VIDEO_MAX_PIXELS = 768 * 28 * 28
VIDEO_TOTAL_PIXELS = 24576 * 28 * 28
FRAME_FACTOR = 2
FPS = 2.0
FPS_MIN_FRAMES = 4
FPS_MAX_FRAMES = 768

OPENAI_CLIP_MEAN = [0.48145466, 0.4578275, 0.40821073]
OPENAI_CLIP_STD = [0.26862954, 0.26130258, 0.27577711]


def round_by_factor(number: float, factor: int) -> int:
    """Return the closest integer to 'number' that is divisible by 'factor'."""
    return round(number / factor) * factor


def ceil_by_factor(number: float, factor: int) -> int:
    """Return the smallest integer greater than or equal to 'number' that is divisible by 'factor'."""
    return math.ceil(number / factor) * factor


def floor_by_factor(number: float, factor: int) -> int:
    """Return the largest integer less than or equal to 'number' that is divisible by 'factor'."""
    return math.floor(number / factor) * factor


def smart_resize(
    height: int,
    width: int,
    factor: int = IMAGE_FACTOR,
    min_pixels: int = MIN_PIXELS,
    max_pixels: int = MAX_PIXELS,
) -> tuple[int, int]:
    """Rescales the image so that the following conditions are met.

    1. Both dimensions (height and width) are divisible by 'factor'.

    2. The total number of pixels is within the range ['min_pixels', 'max_pixels'].

    3. The aspect ratio of the image is maintained as closely as possible.
    """
    if max(height, width) / min(height, width) > MAX_RATIO:
        error_msg = (
            f"absolute aspect ratio must be smaller than {MAX_RATIO}, got {max(height, width) / min(height, width)}"
        )
        raise ValueError(error_msg)
    h_bar = max(factor, round_by_factor(height, factor))
    w_bar = max(factor, round_by_factor(width, factor))
    if h_bar * w_bar > max_pixels:
        beta = math.sqrt((height * width) / max_pixels)
        h_bar = floor_by_factor(height / beta, factor)
        w_bar = floor_by_factor(width / beta, factor)
    elif h_bar * w_bar < min_pixels:
        beta = math.sqrt(min_pixels / (height * width))
        h_bar = ceil_by_factor(height * beta, factor)
        w_bar = ceil_by_factor(width * beta, factor)
    return h_bar, w_bar


def smart_nframes(
    fps: float,
    total_frames: int,
    video_fps: float,
) -> int:
    """Calculate the number of frames for video used for model inputs."""
    min_frames = ceil_by_factor(FPS_MIN_FRAMES, FRAME_FACTOR)
    max_frames = floor_by_factor(min(FPS_MAX_FRAMES, total_frames), FRAME_FACTOR)
    nframes = total_frames / video_fps * fps
    nframes = min(max(nframes, min_frames), max_frames)
    nframes = round_by_factor(nframes, FRAME_FACTOR)

    if not (nframes >= FRAME_FACTOR and nframes <= total_frames):
        error_msg = f"nframes should in interval [{FRAME_FACTOR}, {total_frames}], but got {nframes}."
        raise ValueError(error_msg)
    return nframes


def read_video_cpu(
    video_path: str,
    fps: float,
    num_frames_to_use: int,
    window_range: list[WindowFrameInfo],
) -> tuple[torch.Tensor, list[int]]:
    """Read video using PyAv.

    Args:
        video_path: path to the video support "file://", "http://", "https://" and local path.
        fps: frames per second
        num_frames_to_use: number of frames to use
        window_range: window range

    Returns:
        torch.Tensor: the video tensor with shape (T, C, H, W).

    """
    video_fps = get_avg_frame_rate(video_path)
    idx_list = []
    frame_counts = []
    for window_frame_info in window_range:
        total_frames = window_frame_info.end - window_frame_info.start
        if num_frames_to_use > 0 and num_frames_to_use < total_frames:
            total_frames = num_frames_to_use
            end_frame_idx = window_frame_info.start + num_frames_to_use
        else:
            end_frame_idx = window_frame_info.end
        nframes = smart_nframes(fps, total_frames=total_frames, video_fps=video_fps)
        idx = torch.linspace(window_frame_info.start, end_frame_idx - 1, nframes).round().long().tolist()
        idx_list.extend(idx)
        frame_counts.append(nframes)

    video = decode_video_cpu_frame_ids(video_path, idx_list)
    video = torch.tensor(video).permute(0, 3, 1, 2)  # Convert to TCHW format
    return video, frame_counts


def fetch_video(  # noqa: C901, PLR0911, PLR0913
    video_path: str,
    sampling_fps: float = 2.0,
    window_range: list[WindowFrameInfo] | None = None,
    *,
    do_preprocess: bool = False,
    preprocess_dtype: str = "float32",
    num_frames_to_use: int = 0,
    flip_input: bool = False,
) -> tuple[torch.Tensor, list[int]]:
    """Load and preprocess video frames from a file.

    Args:
        video_path: Path to the video file.
        sampling_fps: Target frames per second for sampling.
        window_range: List of frame windows to extract.
        do_preprocess: Whether to preprocess the frames.
        preprocess_dtype: Data type for preprocessing.
        num_frames_to_use: Number of frames to extract (0 for all).
        flip_input: Whether to flip frames horizontally.

    Returns:
        Tuple of (processed frames tensor, frame indices).

    """
    if window_range is None:
        window_range = []
    video, frame_counts = read_video_cpu(
        video_path,
        sampling_fps,
        num_frames_to_use,
        window_range,
    )
    nframes, _, height, width = video.shape

    max_pixels = max(
        min(VIDEO_MAX_PIXELS, int(VIDEO_TOTAL_PIXELS / nframes * FRAME_FACTOR)),
        int(VIDEO_MIN_PIXELS * 1.05),
    )
    resized_height, resized_width = smart_resize(
        height,
        width,
        factor=IMAGE_FACTOR,
        min_pixels=VIDEO_MIN_PIXELS,
        max_pixels=max_pixels,
    )

    if do_preprocess:
        try:
            desired_dtype = getattr(torch, preprocess_dtype)
        except AttributeError:
            desired_dtype = torch.float32  # Fallback to default dtype

        if preprocess_dtype == "uint8":
            desired_dtype = torch.bfloat16

        resize_norm_transform = v2.Compose(
            [
                v2.Resize(
                    [resized_height, resized_width],
                    interpolation=InterpolationMode.BICUBIC,
                    antialias=True,
                ),
                v2.ToDtype(desired_dtype, scale=True),
                v2.Normalize(mean=OPENAI_CLIP_MEAN, std=OPENAI_CLIP_STD),
            ],
        )
        video = resize_norm_transform(video)
        if flip_input:
            # Flip along the width and height dims
            # TODO: is this a bug?
            video = [torch.flip(frame, dims=[1, 2]) for frame in video]  # type: ignore[assignment]
        if preprocess_dtype == "uint8":
            return video.to(torch.uint8), frame_counts
        return video, frame_counts
    video = transforms.functional.resize(
        video,
        [resized_height, resized_width],
        interpolation=InterpolationMode.BICUBIC,
        antialias=True,
    )
    if flip_input:
        video = [torch.flip(frame, dims=[1, 2]) for frame in video]  # type: ignore[assignment]
    if preprocess_dtype == "float32":
        return video.float(), frame_counts
    if preprocess_dtype == "float16":
        return video.half(), frame_counts
    if preprocess_dtype == "bfloat16":
        return video.to(torch.bfloat16), frame_counts
    if preprocess_dtype == "uint8":
        return video.to(torch.uint8), frame_counts
    return video, frame_counts  # is video a list ? TODO: XXX



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
import queue
from pathlib import Path
from typing import Any

import torch
from loguru import logger

try:
    import cvcuda  # type: ignore[import-untyped]
    import nvcv  # type: ignore[import-untyped]
    import pycuda.driver as cuda  # type: ignore[import-untyped]
    import PyNvVideoCodec as Nvc  # type: ignore[import-untyped]

    pixel_format_to_cvcuda_code = {
        Nvc.Pixel_Format.YUV444: cvcuda.ColorConversion.YUV2RGB,  # type: ignore[import-untyped]
        Nvc.Pixel_Format.NV12: cvcuda.ColorConversion.YUV2RGB_NV12,  # type: ignore[import-untyped]
    }
except (ImportError, RuntimeError):
    logger.warning("PyNvVideoCodec is not installed, some features will be disabled.")
    Nvc = None
    cvcuda = None
    nvcv = None
    cuda = None
    pixel_format_to_cvcuda_code = {}


class FrameExtractionPolicy(enum.Enum):
    """Policy for extracting frames from video, supporting full extraction or FPS-based sampling.

    This enum defines the available strategies for frame extraction from video content.
    """

    full = 0  # Decode video and return all frames
    fps = 1  # Decode video and return frames at a certain sampling fps


class VideoBatchDecoder:
    """GPU-accelerated video decoder that processes video frames in batches.

    This class handles video decoding using NVIDIA hardware acceleration, supporting
    batch processing of frames with color space conversion and resizing capabilities.
    """

    def __init__(  # noqa: PLR0913
        self,
        batch_size: int,
        target_width: int,
        target_height: int,
        device_id: int,
        cuda_ctx: str,  # pyright: ignore[reportAttributeAccessIssue]
        cvcuda_stream: str,  # pyright: ignore[reportAttributeAccessIssue]
    ) -> None:
        """Initialize video batch decoder with GPU acceleration parameters.

        Args:
            batch_size: Number of frames to process in each batch.
            target_width: Target width for decoded frames.
            target_height: Target height for decoded frames.
            device_id: GPU device ID to use.
            cuda_ctx: CUDA context for GPU operations.
            cvcuda_stream: CUDA stream for parallel processing.

        """
        if Nvc is None:
            msg = "PyNvVideoCodec is not available, please install it."
            raise RuntimeError(msg)

        self.batch_size = batch_size
        if self.batch_size <= 0:
            msg = "Batch size should be a valid number."
            raise ValueError(msg)

        self.target_height = target_height
        self.target_width = target_width
        self.device_id = device_id
        self.cuda_ctx = cuda_ctx
        self.cvcuda_stream = cvcuda_stream
        self.decoder: NvVideoDecoder | None = None
        self.torch_YUVtensor = None
        self.cvcuda_RGBtensor = None
        self.nvDemux = None  # pyright: ignore[reportAttributeAccessIssue]
        self.fps = None
        self.prev_batch_size = self.batch_size
        self.input_path: str | None = None

    def get_fps(self) -> int | None:
        """Get the frame rate of the video.

        Returns:
            Frame rate of the video.

        """
        return self.fps

    def __call__(
        self,
        input_path: str,
    ) -> torch.Tensor | None:
        """Process video frames in batches using GPU acceleration.

        Args:
            input_path: Path to the video file to process.

        Returns:
            Processed video frames as a tensor.

        """
        # Check if we need to allocate the decoder for its first use.
        if self.input_path != input_path:
            self.input_path = input_path
            self.decoder = NvVideoDecoder(
                self.input_path,
                self.device_id,
                self.batch_size,
                self.cuda_ctx,
                self.cvcuda_stream,
            )
            self.cvcuda_RGBtensor = None
            self.fps = self.decoder.nvDemux.FrameRate()

        if self.decoder is None:
            msg = "Decoder is not initialized"
            raise RuntimeError(msg)

        # Calculate the target width and height if any one of them is -1
        if self.target_width == -1 or self.target_height == -1:
            width, height = self.decoder.w, self.decoder.h
            minimum_width = 256
            downscale_factor = 1 if width < minimum_width else width // minimum_width

            self.target_width = round(width / downscale_factor)
            self.target_height = round(height / downscale_factor)

        # Get the NHWC YUV tensor from the decoder
        self.torch_YUVtensor = self.decoder.get_next_frames()
        # Check if we are done decoding
        if self.torch_YUVtensor is None:
            return None

        cvcuda_yuv_tensor = cvcuda.as_tensor(self.torch_YUVtensor.cuda(self.device_id), "NHWC")

        # Check the code for the color conversion based in the pixel format
        cvcuda_code = pixel_format_to_cvcuda_code.get(self.decoder.pixelFormat)
        if cvcuda_code is None:
            error_msg = f"Unsupported pixel format: {self.decoder.pixelFormat}"
            raise ValueError(error_msg)

        # Check layout to make sure it is what we expected
        if cvcuda_yuv_tensor.layout != "NHWC":
            error_msg = "Unexpected tensor layout, NHWC expected."
            raise ValueError(error_msg)

        # this may be different than batch size since last frames may not be a multiple of batch size
        actual_batch_size = cvcuda_yuv_tensor.shape[0]

        # Create a CVCUDA tensor for color conversion YUV->RGB
        # Allocate only for the first time or for the last batch.

        if (
            not self.cvcuda_RGBtensor
            or actual_batch_size != self.batch_size
            or actual_batch_size != self.prev_batch_size
        ):
            self.torch_RGBtensor = torch.empty(
                (actual_batch_size, self.decoder.h, self.decoder.w, 3),
                dtype=torch.uint8,
                device=f"cuda:{self.device_id}",
            )
            self.cvcuda_RGBtensor = cvcuda.as_tensor(
                self.torch_RGBtensor.cuda(self.device_id),
                "NHWC",
            )

        # Convert from YUV to RGB. Conversion code is based on the pixel format.
        cvcuda.cvtcolor_into(self.cvcuda_RGBtensor, cvcuda_yuv_tensor, cvcuda_code, stream=self.cvcuda_stream)

        torch_rgb_tensor_resized = torch.empty(
            (self.cvcuda_RGBtensor.shape[0], self.target_height, self.target_width, self.cvcuda_RGBtensor.shape[3]),
            dtype=torch.uint8,
            device=f"cuda:{self.device_id}",
        )
        cvcuda_rgb_tensor_resized = cvcuda.as_tensor(
            torch_rgb_tensor_resized.cuda(self.device_id),
            "NHWC",
        )
        cvcuda.resize_into(
            cvcuda_rgb_tensor_resized,
            self.cvcuda_RGBtensor,
            cvcuda.Interp.LINEAR,
            stream=self.cvcuda_stream,
        )
        self.prev_batch_size = actual_batch_size
        return torch_rgb_tensor_resized


class NvVideoDecoder:
    """Low-level NVIDIA hardware-accelerated video decoder.

    This class provides direct access to NVIDIA's hardware video decoding capabilities,
    handling frame decoding and memory management for video processing pipelines.
    """

    def __init__(
        self,
        enc_file: str,
        device_id: int,
        batch_size: int,
        cuda_ctx: Any,  # noqa: ANN401
        cvcuda_stream: Any,  # noqa: ANN401
    ) -> None:
        """Create instance of HW-accelerated video decoder.

        :param enc_file: Full path to the MP4 file that needs to be decoded.
        :param device_id: id of video card which will be used for decoding & processing.
        :param cuda_ctx: A cuda context object.
        """
        self.device_id = device_id
        self.cuda_ctx = cuda_ctx
        self.input_path = enc_file
        self.cvcuda_stream = cvcuda_stream
        # Demuxer is instantiated only to collect required information about
        # certain video file properties.
        self.nvDemux = Nvc.PyNvDemuxer(self.input_path)  # pyright: ignore[reportAttributeAccessIssue]
        self.nvDec = Nvc.CreateDecoder(  # pyright: ignore[reportAttributeAccessIssue]
            gpuid=self.device_id,
            codec=self.nvDemux.GetNvCodecId(),
            cudacontext=self.cuda_ctx.handle,  # pyright: ignore[reportAttributeAccessIssue]
            cudastream=self.cvcuda_stream.handle,  # pyright: ignore[reportAttributeAccessIssue]
            enableasyncallocations=False,
            usedevicememory=1,
        )
        logger.info("after initializing nvDec")
        self.w, self.h = self.nvDemux.Width(), self.nvDemux.Height()
        self.pixelFormat = self.nvDec.GetPixelFormat()

        self.batch_size = batch_size
        self.decoded_frame_size = None
        self.input_frame_list = queue.Queue()  # type: ignore[X]
        self.nv12_frame_size = None
        self.seq_triggered = False
        self.decoded_frame_cnt = 0
        self.local_frame_index = 0
        self.sent_frame_cnt = 0

    # frame iterator
    def generate_decoded_frames(self) -> list[torch.Tensor]:
        """Generate decoded frames from the video.

        Returns:
            List of decoded frames as tensors.

        """
        for packet in self.nvDemux:
            list_frames = self.nvDec.Decode(packet)
            for decoded_frame in list_frames:
                nvcv_tensor = nvcv.as_tensor(nvcv.as_image(decoded_frame.nvcv_image(), nvcv.Format.U8))
                if nvcv_tensor.layout == "NCHW":
                    nchw_shape = nvcv_tensor.shape
                    nhwc_shape = (nchw_shape[0], nchw_shape[2], nchw_shape[3], nchw_shape[1])
                    torch_nhwc = torch.empty(
                        nhwc_shape,
                        dtype=torch.uint8,
                        device=f"cuda:{self.device_id}",
                    )
                    cvcuda_nhwc = cvcuda.as_tensor(torch_nhwc.cuda(self.device_id), "NHWC")
                    cvcuda.reformat_into(cvcuda_nhwc, nvcv_tensor, stream=self.cvcuda_stream)
                    # Push the decoded frame with the reformatted frame to keep it alive.
                    self.input_frame_list.put(torch_nhwc)
                else:
                    error_msg = "Unexpected tensor layout, NCHW expected."
                    raise ValueError(error_msg)

                self.local_frame_index = self.local_frame_index + 1
                self.decoded_frame_cnt = self.decoded_frame_cnt + 1

            if self.local_frame_index > 0 and self.local_frame_index >= self.batch_size:
                self.sent_frame_cnt = self.sent_frame_cnt + self.local_frame_index
                # `print(f"Frames sent: {self.sent_frame_cnt}, {self.local_frame_index}")`
                sent_frame_list = []
                for _i in range(self.batch_size):
                    sent_frame_list.append(self.input_frame_list.get())
                    self.local_frame_index = self.local_frame_index - 1
                return sent_frame_list

        if self.local_frame_index > 0:
            self.sent_frame_cnt = self.sent_frame_cnt + self.local_frame_index
            # `print(f"Frames sent: {self.sent_frame_cnt}, {self.local_frame_index}")`
            sent_frame_list = []
            for _i in range(self.local_frame_index):
                sent_frame_list.append(self.input_frame_list.get())
                self.local_frame_index = self.local_frame_index - 1
            return sent_frame_list

        return []

    def get_next_frames(self) -> torch.Tensor | None:
        """Get the next frames from the video.

        Returns:
            Next frames from the video as a tensor.

        """
        decoded_frames = self.generate_decoded_frames()
        if len(decoded_frames) == 0:
            return None
        if len(decoded_frames) == 1:  # this case we dont need stack the tensor
            return decoded_frames[0]
        # convert from list of tensors to a single tensor (NHWC)
        # tensorNHWC_torch
        # Sync stream before the decoded frames go out of scope
        return torch.cat(decoded_frames)


def gpu_decode_for_stitching(  # noqa: PLR0913
    device_id: int,
    ctx: str,
    stream: int,
    input_path: Path,
    frame_list: list[int],
    batch_size: int = 1,
) -> list[torch.Tensor]:
    """Decode video frames for stitching using GPU acceleration.

    Args:
        device_id: GPU device ID.
        ctx: CUDA context.
        stream: CUDA stream.
        input_path: Path to the video file to process.
        frame_list: List of frame indices to decode.
        batch_size: Number of frames to process in each batch.

    Returns:
        List of decoded frames as tensors.

    """
    cuda_ctx = ctx
    cvcuda_stream = cvcuda.cuda.as_stream(stream)
    torch_stream = torch.cuda.ExternalStream(cvcuda_stream.handle)
    frames = []

    decoder = VideoBatchDecoder(
        batch_size,
        224,
        224,
        device_id,
        cuda_ctx,
        cvcuda_stream,
    )

    # Loop through all input frames
    frame_idx = 0
    while True:
        # Make sure that cvcuda and torch are using the same stream
        with torch.cuda.stream(torch_stream):  # pyright: ignore[reportArgumentType]
            batch = decoder(input_path.as_posix())
            if batch is None:
                break  # No more frames to decode

            actual_batch_size = batch.shape[0]
            batch = torch.as_tensor(batch.cuda(), device=f"cuda:{device_id}")

            for i in range(frame_idx, frame_idx + actual_batch_size):
                if i in frame_list:
                    # There can be cases where the clip is so small that boundary frames fall outside
                    # or end up being the same. In such case, simply checking if "i in frame_list" won't work.
                    # I am not sure if we want to discard such clips or keep them or what to do about the boundary.
                    # So, this is a pure hack to make the pipeline work.

                    for _j in range(frame_list.count(i)):
                        # use list.extend
                        frames.append(batch[i - frame_idx, :, :, :])

            frame_idx += actual_batch_size

    # No sync required here because it's in the same thread
    return frames


class PyNvcFrameExtractor:
    """High-level frame extraction interface using PyNvVideoCodec.

    This class provides a simplified interface for extracting frames from videos using
    hardware acceleration, supporting both full extraction and FPS-based sampling.
    """

    def __init__(
        self,
        width: int,
        height: int,
        batch_size: int,
    ) -> None:
        """Initialize the PyNvcFrameExtractor.

        Args:
            width: Width of the video frames.
            height: Height of the video frames.
            batch_size: Number of frames to process in each batch.

        """
        device_id = 0
        cuda_device = cuda.Device(device_id)  # type: ignore[X]
        cuda_ctx = cuda_device.retain_primary_context()
        cuda_ctx.push()
        cvcuda_stream = cvcuda.Stream()
        # `cvcuda_stream = cvcuda.cuda.as_stream(cvcuda_stream)`
        self.torch_stream = torch.cuda.ExternalStream(cvcuda_stream.handle)
        cuda_ctx.pop()

        self.decoder = VideoBatchDecoder(
            batch_size,
            width,
            height,
            device_id,
            cuda_ctx,
            cvcuda_stream,
        )

    def __call__(
        self,
        input_path: Path,
        extraction_policy: FrameExtractionPolicy = FrameExtractionPolicy.full,
        sampling_fps: int = 1,
    ) -> torch.Tensor:
        """Extract frames from the video.

        Args:
            input_path: Path to the video file to process.
            extraction_policy: Policy for extracting frames.
            sampling_fps: Sampling rate for FPS-based extraction.

        Returns:
            List of decoded frames as tensors.

        """
        tensor_list = []
        video_fps: int | None = None
        # Loop through all input frames
        while True:
            # Make sure that cvcuda and torch are using the same stream
            with torch.cuda.stream(self.torch_stream):  # pyright: ignore[reportArgumentType]
                batch = self.decoder(input_path.as_posix())
                if batch is None:
                    break  # No more frames to decode
                if extraction_policy == FrameExtractionPolicy.fps:
                    if video_fps is None:
                        video_fps = self.decoder.get_fps()
                        if video_fps is None:
                            msg = "Unable to get video FPS"
                            raise RuntimeError(msg)
                    batch = batch[:: int(video_fps / sampling_fps)]
                tensor_list.append(batch)
        return torch.cat(tensor_list, dim=0)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field
from typing import Any

from nemo_curator.tasks import DocumentBatch

from .base import BaseWriter


@dataclass
class JsonlWriter(BaseWriter):
    """Writer that writes a DocumentBatch to a JSONL file."""

    # Additional kwargs for pandas.DataFrame.to_json
    file_extension: str = "jsonl"
    write_kwargs: dict[str, Any] = field(default_factory=dict)
    _name: str = "jsonl_writer"

    def write_data(self, task: DocumentBatch, file_path: str) -> None:
        """Write data to JSONL file using pandas DataFrame.to_json."""
        df = task.to_pandas()  # Convert to pandas DataFrame if needed
        # Filter fields if specified
        if self.fields is not None:
            df = df[self.fields]

        # Build kwargs for to_json with explicit options
        write_kwargs = {
            "lines": True,
            "orient": "records",
        }

        # Add any additional kwargs, allowing them to override defaults
        write_kwargs.update(self.write_kwargs)
        df.to_json(file_path, **write_kwargs)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Any

from loguru import logger
from transformers import AutoTokenizer

from nemo_curator.utils.hf_download_utils import download_model_from_hf

try:
    from vllm import LLM, SamplingParams

    VLLM_AVAILABLE = True
except ImportError:
    VLLM_AVAILABLE = False

    # Create dummy classes for type hints when vllm is not available
    class LLM:
        pass

    class SamplingParams:
        pass


from nemo_curator.models.base import ModelInterface

_QWEN_LM_MODEL_ID = "Qwen/Qwen2.5-14B-Instruct"
_QWEN_LM_MODEL_REVISION = "cf98f3b"


class QwenLM(ModelInterface):
    """Qwen language model."""

    def model_id_names(self) -> list[str]:
        return [_QWEN_LM_MODEL_ID]

    def __init__(self, model_dir: str, caption_batch_size: int, fp8: bool, max_output_tokens: int):
        self.model_dir = model_dir
        self.caption_batch_size = caption_batch_size
        self.fp8 = fp8
        self.max_output_tokens = max_output_tokens

    def setup(self) -> None:
        if not VLLM_AVAILABLE:
            msg = "vllm is required for QwenLM model but is not installed. Please install vllm: pip install vllm"
            raise ImportError(msg)

        self.weight_file = str(Path(self.model_dir) / _QWEN_LM_MODEL_ID)
        self.llm = LLM(
            model=self.weight_file,
            quantization="fp8" if self.fp8 else None,
            enforce_eager=False,
        )
        self.sampling_params = SamplingParams(
            temperature=0.1,
            top_p=0.001,
            repetition_penalty=1.05,
            max_tokens=self.max_output_tokens,
            stop_token_ids=[],
        )
        self.tokenizer = AutoTokenizer.from_pretrained(self.weight_file)

    def generate(self, inputs: list[dict[str, Any]]) -> list[str]:
        formatted_inputs = self.tokenizer.apply_chat_template(inputs, tokenize=False, add_generation_prompt=True)
        results = self.llm.generate(formatted_inputs, sampling_params=self.sampling_params)
        return [result.outputs[0].text for result in results]

    @classmethod
    def download_weights_on_node(cls, model_dir: str) -> None:
        """Download the weights for the QwenLM model on the node."""
        model_dir_path = Path(model_dir) / _QWEN_LM_MODEL_ID
        model_dir_path.mkdir(parents=True, exist_ok=True)
        if model_dir_path.exists() and any(model_dir_path.glob("*.safetensors")):
            return
        download_model_from_hf(
            model_id=_QWEN_LM_MODEL_ID,
            local_dir=model_dir_path,
            revision=_QWEN_LM_MODEL_REVISION,
        )
        logger.info(f"QwenLM weights downloaded to: {model_dir_path}")



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import socket
import subprocess
from typing import TYPE_CHECKING

import ray
from loguru import logger

from nemo_curator.core.constants import (
    DEFAULT_RAY_AUTOSCALER_METRIC_PORT,
    DEFAULT_RAY_DASHBOARD_METRIC_PORT,
    DEFAULT_RAY_MAX_WORKER_PORT,
    DEFAULT_RAY_MIN_WORKER_PORT,
)

if TYPE_CHECKING:
    import loguru


def get_free_port(start_port: int, get_next_free_port: bool = True) -> int:
    """Checks if start_port is free.
    If not, it will get the next free port starting from start_port if get_next_free_port is True.
    Else, it will raise an error if the free port is not equal to start_port.
    """
    for port in range(start_port, 65535):
        if port >= DEFAULT_RAY_MIN_WORKER_PORT and port <= DEFAULT_RAY_MAX_WORKER_PORT:
            continue
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            # SO_REUSEADDR to avoid TIME_WAIT issues on some OSes
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            try:
                s.bind(("localhost", port))
                # If bind succeeds, port is free
                return port  # noqa: TRY300
            except OSError:
                if not get_next_free_port and port == start_port:
                    msg = f"Port {start_port} is already in use. Please provide a different port."
                    raise RuntimeError(msg)  # noqa: B904
                continue
    msg = f"No free port found between {start_port} and 65535"
    raise RuntimeError(msg)


def _logger_custom_serializer(
    _: "loguru.Logger",
) -> None:
    return None


def _logger_custom_deserializer(
    _: None,
) -> "loguru.Logger":
    # Initialize a default logger
    return logger


def init_cluster(  # noqa: PLR0913
    ray_port: int,
    ray_temp_dir: str,
    ray_dashboard_port: int,
    ray_metrics_port: int,
    ray_client_server_port: int,
    ray_dashboard_host: str,
    num_gpus: int | None = None,
    num_cpus: int | None = None,
    enable_object_spilling: bool = False,
    block: bool = True,
    ip_address: str | None = None,
) -> subprocess.Popen:
    """Initialize a new local Ray cluster or connects to an existing one."""
    # Turn off serization for loguru. This is needed as loguru is not serializable in general.
    ray.util.register_serializer(
        logger.__class__,
        serializer=_logger_custom_serializer,
        deserializer=_logger_custom_deserializer,
    )

    ip_address = ip_address or socket.gethostbyname(socket.gethostname())
    ray_command = ["ray", "start", "--head"]
    ray_command.extend(["--node-ip-address", ip_address])
    ray_command.extend(["--port", str(ray_port)])
    ray_command.extend(["--metrics-export-port", str(ray_metrics_port)])
    ray_command.extend(["--dashboard-host", ray_dashboard_host])
    ray_command.extend(["--dashboard-port", str(ray_dashboard_port)])
    ray_command.extend(["--ray-client-server-port", str(ray_client_server_port)])
    ray_command.extend(["--temp-dir", ray_temp_dir])
    ray_command.extend(["--disable-usage-stats"])
    if enable_object_spilling:
        ray_command.extend(
            [
                "--system-config",
                '{"local_fs_capacity_threshold": 0.95, "object_spilling_config": "{ "type": "filesystem", "params": {"directory_path": "/tmp/ray_spill", "buffer_size": 1000000 } }"}',
            ]
        )
    if num_gpus:
        ray_command.extend(["--num-gpus", str(num_gpus)])
    if num_cpus:
        ray_command.extend(["--num-cpus", str(num_cpus)])
    if block:
        ray_command.extend(["--block"])

    # We need to set these env vars to ensure that metrics of ray dashboard and autoscaler are available for various different clusters.
    os.environ["DASHBOARD_METRIC_PORT"] = str(get_free_port(DEFAULT_RAY_DASHBOARD_METRIC_PORT))
    os.environ["AUTOSCALER_METRIC_PORT"] = str(get_free_port(DEFAULT_RAY_AUTOSCALER_METRIC_PORT))

    # We set some env vars for Xenna here. This is only used for Xenna clusters.
    os.environ["XENNA_RAY_METRICS_PORT"] = str(ray_metrics_port)
    os.environ["XENNA_RESPECT_CUDA_VISIBLE_DEVICES"] = "1"

    proc = subprocess.Popen(ray_command, shell=False)  # noqa: S603
    logger.info(f"Ray start command: {' '.join(ray_command)}")
    return proc



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field
from typing import Any

import nemo.collections.asr as nemo_asr
import torch

from nemo_curator.backends.base import NodeInfo, WorkerMetadata
from nemo_curator.stages.base import ProcessingStage
from nemo_curator.stages.resources import Resources
from nemo_curator.tasks import AudioBatch, DocumentBatch, FileGroupTask


@dataclass
class InferenceAsrNemoStage(ProcessingStage[FileGroupTask | DocumentBatch | AudioBatch, AudioBatch]):
    """Stage that do speech recognition inference using NeMo model.

    Args:
        model_name (str): name of the speech recognition NeMo model. See full list at https://docs.nvidia.com/nemo-framework/user-guide/latest/nemotoolkit/asr/all_chkpt.html
        asr_model (Any): ASR model object. Defaults to None
        filepath_key (str): which key of the data object should be used to find the path to audiofile. Defaults to “audio_filepath”
        pred_text_key (str): key is used to identify the field containing the predicted transcription associated with a particular audio sample. Defaults to “pred_text”
        name (str): Stage name. Defaults to "ASR_inference"
    """

    model_name: str
    asr_model: Any | None = None
    filepath_key: str = "audio_filepath"
    pred_text_key: str = "pred_text"
    _name: str = "ASR_inference"
    _batch_size: int = 16
    _resources: Resources = field(default_factory=lambda: Resources(cpus=1.0))

    def check_cuda(self) -> torch.device:
        return torch.device("cuda") if self.resources.gpus > 0 else torch.device("cpu")

    def setup_on_node(self, _node_info: NodeInfo | None = None, _worker_metadata: WorkerMetadata = None) -> None:
        self.setup()

    def setup(self, _worker_metadata: WorkerMetadata = None) -> None:
        """Initialise heavy object self.asr_model: nemo_asr.models.ASRModel"""
        if not self.asr_model:
            try:
                map_location = self.check_cuda()
                self.asr_model = nemo_asr.models.ASRModel.from_pretrained(
                    model_name=self.model_name, map_location=map_location
                )
            except Exception as e:
                msg = f"Failed to download {self.model_name}"
                raise RuntimeError(msg) from e

    def inputs(self) -> tuple[list[str], list[str]]:
        """Define the input attributes required by this stage.

        Returns:
            Tuple of (top_level_attrs, data_attrs) where:
            - top_level_attrs: ["data"] - requires FileGroupTask.data to be populated
        """
        return ["data"], []

    def outputs(self) -> tuple[list[str], list[str]]:
        """Define the output attributes produced by this stage.

        Returns:
            Tuple of (top_level_attrs, data_attrs) where:
            - top_level_attrs: ["data"] - populates FileGroupTask.data
            - data_attrs: [self.filepath_key, self.pred_text_key] - audiofile path and predicted text.
        """
        return ["data"], [self.filepath_key, self.pred_text_key]

    def transcribe(self, files: list[str]) -> list[str]:
        """Run inference for speech recognition model
         Args:
            files: list of audio file paths.

        Returns:
            list of predicted texts.
        """

        outputs = self.asr_model.transcribe(files)

        # Tuple (hyps, all_hyps) noqa: ERA001
        if isinstance(outputs, tuple):
            outputs = outputs[0]

        # list[list[Hypothesis]] noqa: ERA001
        if outputs and isinstance(outputs[0], list):
            if outputs[0] and hasattr(outputs[0][0], "text"):
                return [inner[0].text for inner in outputs]
            return [inner[0] for inner in outputs]

        return [output.text for output in outputs]

    def process(self, task: FileGroupTask | DocumentBatch | AudioBatch) -> AudioBatch:
        """Process a audio task by reading audio file and do ASR inference.


        Args:
            tasks: List of FileGroupTask containing a path to audop file for inference.

        Returns:
            List of SpeechObject with self.filepath_key .
            If errors occur, the task is returned with error information stored.
        """
        files = []
        audio_items = []

        if not self.validate_input(task):
            msg = f"Task {task!s} failed validation for stage {self}"
            raise ValueError(msg)
        if isinstance(task, FileGroupTask):
            files = [task.data[0]]
        elif isinstance(task, DocumentBatch):
            files = list(task.data[self.filepath_key])
        elif isinstance(task, AudioBatch):
            files = [item[self.filepath_key] for item in task.data]
        else:
            raise TypeError(str(task))

        outputs = self.transcribe(files)

        for i, text in enumerate(outputs):
            entry = task.data[i]
            file_path = files[i]

            if isinstance(entry, dict):
                item = entry
                item[self.pred_text_key] = text
            else:
                item = {self.filepath_key: file_path, self.pred_text_key: text}
            audio_items.append(item)

        return AudioBatch(
            task_id=f"task_id_{self.model_name}",
            dataset_name=f"{self.model_name}_inference",
            filepath_key=self.filepath_key,
            data=audio_items,
        )



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gzip
import os
import re
import tarfile
import tempfile
from collections.abc import Iterator
from typing import Any

from loguru import logger

from nemo_curator.stages.text.download import DocumentIterator
from nemo_curator.utils.file_utils import get_all_file_paths_under, tar_safe_extract

# The iterator and extractor code are in large part taken
# from the Red-Pajama repo
# https://github.com/togethercomputer/RedPajama-Data/tree/main/data_prep/arxiv


class ArxivIterator(DocumentIterator):
    """Processes downloaded Arxiv files and extracts article content."""

    def __init__(self, log_frequency: int = 1000):
        super().__init__()
        self._log_frequency = log_frequency
        self._counter = 0

    def _tex_proj_loader(self, file_or_dir_path: str) -> list[str] | None:
        r"""function to load the tex files from a tar file or a gzip file. The
        function will return a tuple containing a list of tex files and the
        timestamp of the project.

        @param file_or_dir_path: path to the tar file or the gzip file

        @return: tuple containing a list of tex files and the timestamp of the
            project
        """
        files_and_content = []

        try:
            # if it is a directory, open it as a tarfile
            with tarfile.open(file_or_dir_path) as sub_tf:
                for member in sub_tf.getmembers():
                    if member.name.endswith(".tex"):
                        file_content = sub_tf.extractfile(member).read()

                        try:
                            file_content = file_content.decode("utf-8")
                        except UnicodeDecodeError:
                            logger.debug(f"UnicodeDecodeError: {file_or_dir_path}")
                            return None

                        files_and_content.append(file_content)

        except tarfile.ReadError:
            # otherwise we try opening it as a gzip file
            try:
                with gzip.open(file_or_dir_path, "rb") as gz:
                    file_content = gz.read()
            except Exception as e:  # noqa: BLE001
                # all fails, we skip this file
                logger.debug(f"[ERROR] {e}: {file_or_dir_path}")
                return None

            try:
                file_content = file_content.decode("utf-8")
            except UnicodeDecodeError:
                logger.debug(f"UnicodeDecodeError: {file_or_dir_path}")
                return None

            files_and_content.append(file_content)

        except Exception as e:  # noqa: BLE001
            logger.debug(f"[ERROR] {e}: {file_or_dir_path}")
            return None

        return files_and_content

    def _format_arxiv_id(self, arxiv_id: str) -> str:
        r"""this function brings the raw arxiv-id into a format compliant with the
        specification from arxiv. This is used to create the url to the arxiv
        abstract page.

        - Format prior to March 2007:
            <archive>/YYMMNNN where N is a 3-digit number
        - Format after March 2007: <archive>/YYMM.NNNNN where N is a
          5 (or 6)-digit number

        References: https://info.arxiv.org/help/arxiv_identifier.html

        @param arxiv_id: raw arxiv id which can be in one of the
                         following formats:
                         - <archive><YY><MM><NNN>
                         - <YY><MM><NNNNN|NNNNNN>

        @return: formatted arxiv id
        """
        match = re.search(r"^([a-zA-Z-]*)([\d\.]+)$", arxiv_id)

        if match is None:
            msg = f"Invalid arxiv id: {arxiv_id}"
            raise ValueError(msg)

        if match.group(1) == "":
            return match.group(2)

        return f"{match.group(1)}/{match.group(2)}"

    def iterate(self, file_path: str) -> Iterator[dict[str, Any]]:
        self._counter = 0
        download_dir = os.path.split(file_path)[0]
        bname = os.path.split(file_path)[-1]

        with tempfile.TemporaryDirectory(dir=download_dir) as tmpdir, tarfile.open(file_path) as tf:
            # Use safe extraction instead of extractall to prevent path traversal attacks
            tar_safe_extract(tf, tmpdir)
            for _i, item in enumerate(get_all_file_paths_under(tmpdir, recurse_subdirectories=True)):
                if self._counter > 0 and self._counter % self._log_frequency == 0:
                    logger.info(f"Extracted {self._counter} papers from {file_path}")
                self._counter += 1

                tex_files = self._tex_proj_loader(item)
                arxiv_id = os.path.splitext(os.path.split(item)[-1])[0]

                # get the arxiv id in the correct format
                try:
                    clean_arxiv_id = self._format_arxiv_id(arxiv_id)
                except Exception as e:  # noqa: BLE001
                    logger.debug(f"[WARNING] failed to format arxiv id {arxiv_id}; exception={e}")
                    clean_arxiv_id = arxiv_id

                if tex_files is None:
                    continue

                yield {"id": clean_arxiv_id, "source_id": f"{bname}", "content": tex_files}

    def output_columns(self) -> list[str]:
        return ["id", "source_id", "content"]
